import os
import random
import time
from pathlib import Path
from queue import Queue

import gevent
import jsonlines
import locust
import pandas as pd
from openai import OpenAI
from tqdm import tqdm

RANDOM_SEED = 42


def _get_model_name(model_id: str) -> str:
    segments = model_id.split("/")

    # Get first 5 characters of the snapshot id to keep the name small.
    snapshot = segments[-1][:5]
    model_segment_key = "models--"
    for s in segments:
        if model_segment_key in s:
            model_name = s[len(model_segment_key) :]
            return f"{model_name}_{snapshot}"
    # If the model_id is not in the HF format (ex. it's a path to a custom model),
    # return up to the last 20 characters.
    # We replace slashes in the directory path with underscores since the model
    # name is used as part of a filename.
    return model_id.replace("/", "_")[-20:]


def _get_metrics(
    total_input_tokens: int,
    total_output_tokens: int,
    delta_time: float,
    requests_completed: int,
):
    input_tokens_per_second = total_input_tokens / delta_time
    output_tokens_per_second = total_output_tokens / delta_time
    total_tokens_per_second = (total_input_tokens + total_output_tokens) / delta_time
    elapsed_time = time.strftime("%H:%M:%S", time.gmtime(delta_time))
    metrics = {
        "Total tok/s": round(total_tokens_per_second, 2),
        "Total tokens": total_input_tokens + total_output_tokens,
        "Input tok/s": round(input_tokens_per_second, 2),
        "Input tokens": total_input_tokens,
        "Output tok/s": round(output_tokens_per_second, 2),
        "Output tokens": total_output_tokens,
        "Requests/s": round(requests_completed / delta_time, 2),
        "Request count": requests_completed,
        "Elapsed Time": elapsed_time,
        "Elapsed Time (seconds)": round(delta_time, 2),
    }
    return metrics


def main():
    """Run parallelized inference against a vLLM server."""
    locust.log.setup_logging("INFO")
    random.seed(RANDOM_SEED)

    IP = os.environ["THIS_IP_ADDRESS"]
    openai_api_base = f"http://{IP}:8000/v1"
    client = OpenAI(
        base_url=openai_api_base,
    )
    models = client.models.list()
    MODEL = models.data[0].id
    MODEL_NAME = _get_model_name(MODEL)

    NUM_WORKERS = int(os.environ["OUMI_VLLM_NUM_WORKERS"])
    SPAWN_RATE = int(os.environ["OUMI_VLLM_WORKERS_SPAWNED_PER_SECOND"])
    print(f"Num workers: {NUM_WORKERS}")
    JOB_NUMBER = os.environ["JOB_NUMBER"]
    INPUT_FILEPATH = Path(os.environ["OUMI_VLLM_INPUT_FILEPATH"])
    OUTPUT_DIR = Path(os.environ["OUMI_VLLM_OUTPUT_DIR"])
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    TIMESTR = time.strftime("%Y%m%d_%H%M%S")
    OUTPUT_FILENAME = f"{JOB_NUMBER}_vllm_output_{TIMESTR}_{MODEL_NAME}.jsonl"
    METRIC_FILENAME = f"{JOB_NUMBER}_vllm_metrics_{TIMESTR}_{MODEL_NAME}.jsonl"
    OUTPUT_FILEPATH = OUTPUT_DIR / OUTPUT_FILENAME
    METRIC_FILEPATH = OUTPUT_DIR / METRIC_FILENAME
    REQUEST_RETRIES = 3
    print(f"Input file is {INPUT_FILEPATH}")
    print(f"Files will be output to {OUTPUT_DIR}")

    if not INPUT_FILEPATH.is_file():
        raise FileNotFoundError(f"Input file not found: {INPUT_FILEPATH}")
    json_objects = pd.read_json(INPUT_FILEPATH, lines=True)
    ALL_MESSAGES = json_objects["messages"].to_list()

    global output_queue
    output_queue = Queue()
    global input_queue
    input_queue = Queue()

    for i in range(len(ALL_MESSAGES)):
        input_queue.put(i)

    global REQUEST_TIMES
    REQUEST_TIMES = [(0.0, 0.0) for _ in range(len(ALL_MESSAGES))]
    global failed_request_counts
    failed_request_counts = {}

    class VllmWorker(locust.FastHttpUser):
        host = openai_api_base
        network_timeout = 300
        connection_timeout = 300
        max_retries = 3

        @locust.task
        def run_inference(self):
            """Runs inference by pulling indices from queue shared by all workers."""
            global output_queue
            global input_queue
            global REQUEST_TIMES
            global failed_request_counts

            while True:
                index = input_queue.get()
                input_queue.task_done()
                messages = ALL_MESSAGES[index]
                payload = {"model": MODEL, "messages": messages, "seed": RANDOM_SEED}
                try:
                    request_sent_time = time.perf_counter()
                    response = self.client.post(url="/chat/completions", json=payload)
                    request_complete_time = time.perf_counter()
                    response_dict = response.json()
                    response_message = response_dict["choices"][0]["message"]["content"]
                    num_input_tokens = response_dict["usage"]["prompt_tokens"]
                    num_output_tokens = response_dict["usage"]["completion_tokens"]
                    output_queue.put(
                        (index, response_message, num_input_tokens, num_output_tokens)
                    )
                    REQUEST_TIMES[index] = (request_sent_time, request_complete_time)
                except Exception as e:
                    print(e)
                    if index not in failed_request_counts:
                        failed_request_counts[index] = 0

                    failed_request_counts[index] += 1
                    if failed_request_counts[index] >= REQUEST_RETRIES:
                        print(f"Failed for request at index {index}, skipping...")
                        continue
                    input_queue.put(index)

    # setup Environment and Runner
    env = locust.env.Environment(user_classes=[VllmWorker], events=locust.events)
    runner = env.create_local_runner()

    # execute init event handlers (only really needed if you have registered any)
    env.events.init.fire(environment=env, runner=runner)

    # start a greenlet that periodically outputs the current stats
    gevent.spawn(locust.stats.stats_printer(env.stats))

    # start a greenlet that save current stats to history
    gevent.spawn(locust.stats.stats_history, env.runner)

    # start the test
    runner.start(NUM_WORKERS, spawn_rate=SPAWN_RATE)

    print("Waiting for inference to finish.")
    start = time.perf_counter()
    pbar = tqdm(total=len(ALL_MESSAGES))
    total_input_tokens = 0
    total_output_tokens = 0
    requests_completed = 0
    while requests_completed < len(ALL_MESSAGES):
        queue_item = output_queue.get()
        index, response_message, num_input_tokens, num_output_tokens = queue_item
        output_queue.task_done()
        messages = ALL_MESSAGES[index]
        total_input_tokens += num_input_tokens
        total_output_tokens += num_output_tokens

        messages.append({"role": "assistant", "content": response_message})
        requests_completed += 1
        pbar.update()

        with jsonlines.open(OUTPUT_FILEPATH, mode="a") as writer:
            request_sent_time, request_complete_time = REQUEST_TIMES[index]
            elapsed_time = request_complete_time - request_sent_time
            json_obj = {
                "messages": messages,
                "original_index": index,
                "num_input_tokens": num_input_tokens,
                "num_output_tokens": num_output_tokens,
                "elapsed_time": elapsed_time,
                "request_sent_time": request_sent_time,
                "request_completed_time": request_complete_time,
            }
            writer.write(json_obj)

        if requests_completed % 50 == 0:
            metrics = _get_metrics(
                total_input_tokens,
                total_output_tokens,
                time.perf_counter() - start,
                requests_completed,
            )

            with jsonlines.open(METRIC_FILEPATH, "a") as metric_writer:
                metric_writer.write(metrics)

    average_request_completion_time = sum([x[1] - x[0] for x in REQUEST_TIMES]) / len(
        REQUEST_TIMES
    )
    metrics = _get_metrics(
        total_input_tokens,
        total_output_tokens,
        time.perf_counter() - start,
        requests_completed,
    )
    metrics["Avg Request Completion Time (seconds)"] = round(
        average_request_completion_time, 2
    )

    with jsonlines.open(METRIC_FILEPATH, "a") as metric_writer:
        metric_writer.write(metrics)

    print("All responses written to file.")
    runner.quit()

    print("Joining runners...")
    # wait for the greenlets
    runner.greenlet.join()
    print("OUMI INFERENCE JOB DONE")


if __name__ == "__main__":
    main()



from unittest.mock import patch

import torch

from oumi.builders.callbacks import build_training_callbacks
from oumi.core.callbacks.mfu_callback import MfuTrainerCallback
from oumi.core.callbacks.nan_inf_detection_callback import NanInfDetectionCallback
from oumi.core.callbacks.telemetry_callback import TelemetryCallback
from oumi.core.configs import (
    TrainingConfig,
)


def test_build_training_callbacks_no_performance_metrics():
    config = TrainingConfig()
    model = torch.nn.Module()
    result = build_training_callbacks(config, model, None)
    assert result == []


def test_build_training_callbacks_mfu_callback():
    config = TrainingConfig()
    config.training.include_performance_metrics = True
    config.data.train.pack = True
    config.model.model_max_length = 128
    model = torch.nn.Sequential(torch.nn.Linear(10, 10))
    with patch("torch.cuda.is_available", return_value=True):
        with patch("torch.cuda.get_device_name", return_value="NVIDIA A100-PCIE-40GB"):
            result = build_training_callbacks(config, model, None)
    assert len(result) == 3
    assert isinstance(result[0], MfuTrainerCallback)
    assert isinstance(result[1], NanInfDetectionCallback)
    assert isinstance(result[2], TelemetryCallback)


@patch("oumi.utils.logging.logger.warning")
def test_build_training_callbacks_no_cuda(mock_logger_warning):
    config = TrainingConfig()
    config.training.include_performance_metrics = True
    model = torch.nn.Sequential(torch.nn.Linear(10, 10))
    with patch("torch.cuda.is_available", return_value=False):
        result = build_training_callbacks(config, model, None)
    assert len(result) == 2
    assert isinstance(result[0], NanInfDetectionCallback)
    assert isinstance(result[1], TelemetryCallback)
    mock_logger_warning.assert_called_with(
        "MFU logging is only supported on GPU. Skipping MFU callbacks."
    )


@patch("oumi.utils.logging.logger.warning")
def test_build_training_callbacks_peft(mock_logger_warning):
    config = TrainingConfig()
    config.training.include_performance_metrics = True
    config.training.use_peft = True
    model = torch.nn.Sequential(torch.nn.Linear(10, 10))
    with patch("torch.cuda.is_available", return_value=True):
        result = build_training_callbacks(config, model, None)
    assert len(result) == 2
    assert isinstance(result[0], NanInfDetectionCallback)
    assert isinstance(result[1], TelemetryCallback)
    mock_logger_warning.assert_called_with(
        "MFU logging is not supported for PEFT. Skipping MFU callbacks."
    )


@patch("oumi.utils.logging.logger.warning")
def test_build_training_callbacks_no_pack(mock_logger_warning):
    config = TrainingConfig()
    config.training.include_performance_metrics = True
    model = torch.nn.Sequential(torch.nn.Linear(10, 10))
    with patch("torch.cuda.is_available", return_value=True):
        result = build_training_callbacks(config, model, None)
    assert len(result) == 2
    assert isinstance(result[0], NanInfDetectionCallback)
    assert isinstance(result[1], TelemetryCallback)
    mock_logger_warning.assert_called_with(
        "MFU logging requires packed datasets. Skipping MFU callbacks."
    )


@patch("oumi.utils.logging.logger.warning")
def test_build_training_callbacks_unknown_device_name(mock_logger_warning):
    config = TrainingConfig()
    config.training.include_performance_metrics = True
    config.data.train.pack = True
    model = torch.nn.Sequential(torch.nn.Linear(10, 10))
    with patch("torch.cuda.is_available", return_value=True):
        with patch("torch.cuda.get_device_name", return_value="Foo"):
            result = build_training_callbacks(config, model, None)
    assert len(result) == 2
    assert isinstance(result[0], NanInfDetectionCallback)
    assert isinstance(result[1], TelemetryCallback)
    mock_logger_warning.assert_called_with(
        "MFU logging is currently not supported for device Foo. Skipping MFU callbacks."
    )


@patch("oumi.utils.logging.logger.warning")
def test_build_training_callbacks_no_model_max_length(mock_logger_warning):
    config = TrainingConfig()
    config.training.include_performance_metrics = True
    config.data.train.pack = True
    model = torch.nn.Sequential(torch.nn.Linear(10, 10))
    with patch("torch.cuda.is_available", return_value=True):
        with patch("torch.cuda.get_device_name", return_value="NVIDIA A100-PCIE-40GB"):
            result = build_training_callbacks(config, model, None)
    assert len(result) == 2
    assert isinstance(result[0], NanInfDetectionCallback)
    assert isinstance(result[1], TelemetryCallback)
    mock_logger_warning.assert_called_with(
        "model_max_length must be set to log MFU performance information."
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.types.conversation import Conversation, Role

_DEFAULT_INSTRUCTION_FIELD_NAME = "instruction"
_DEFAULT_OUTPUT_FIELD_NAME = "output"


def conversation_to_alpaca_format(
    conversation: Conversation,
    instruction_field_name: str = _DEFAULT_INSTRUCTION_FIELD_NAME,
    output_field_name: str = _DEFAULT_OUTPUT_FIELD_NAME,
) -> dict:
    """Converts an Oumi `Conversation` to Alpaca format.

    Converts an Oumi single-turn conversation to a dictionary with keys `instruction`
    and `output`. If the first message is a System Instruction, it is ignored. Any
    fields in the conversation's metadata are also retained as dict entries.

    Alpaca format (see: https://huggingface.co/datasets/tatsu-lab/alpaca_eval):
        {
            "instruction": <`str` prompt/request to a Large Language Model (LLM)>,
            "output": <`str` response by a Large Language Model (LLM)>,
        }
    """
    # Ensure the number of messages is correct.
    if len(conversation.messages) not in (2, 3):
        raise ValueError("Only single-turn conversations are currently supported.")

    # Ensure that the first message is an SI (if we have 3 messages).
    if len(conversation.messages) == 3:
        if conversation.messages[0].role != Role.SYSTEM:
            raise ValueError(
                f"Role of first message is `{conversation.messages[0].role}`, "
                "while `Role.SYSTEM` is expected for conversations of 3 messages."
            )

    # Extract the instruction and output.
    instruction = conversation.messages[-2]
    output = conversation.messages[-1]

    # Ensure that the roles for {instruction, output} are correct.
    if instruction.role != Role.USER:
        raise ValueError("Role of `instruction` should be `Role.USER`")
    if output.role != Role.ASSISTANT:
        raise ValueError("Role of `output` should be `Role.ASSISTANT`")

    # Extract metadata to add.
    metadata = {}
    if conversation.conversation_id is not None:
        metadata["conversation_id"] = conversation.conversation_id
    metadata.update(conversation.metadata)
    metadata.pop(instruction_field_name, None)
    metadata.pop(output_field_name, None)

    # Create a dictionary with the instruction, output, metadata.
    conversations_dict = {
        instruction_field_name: instruction.content,
        output_field_name: output.content,
    }
    conversations_dict.update(metadata)
    return conversations_dict


def conversations_to_alpaca_format(
    conversations: list[Conversation],
    instruction_field_name: str = _DEFAULT_INSTRUCTION_FIELD_NAME,
    output_field_name: str = _DEFAULT_OUTPUT_FIELD_NAME,
) -> list[dict]:
    """Converts a list of conversations to Alpaca format (list of dictionaries)."""
    return [
        conversation_to_alpaca_format(
            conversation=conversation,
            instruction_field_name=instruction_field_name,
            output_field_name=output_field_name,
        )
        for conversation in conversations
    ]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic class for using HuggingFace datasets with messages column.

Allows users to specify the messages column at the config level.
"""

from typing import Union

import pandas as pd

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("HuggingFaceDataset")
class HuggingFaceDataset(BaseSftDataset):
    """Converts HuggingFace Datasets with messages to Oumi Message format.

    Example::

        dataset = HuggingFaceDataset(
            hf_dataset_path="oumi-ai/oumi-synthetic-document-claims",
            message_column="messages"
        )
    """

    def __init__(
        self,
        *,
        hf_dataset_path: str = "",
        messages_column: str = "messages",
        exclude_final_assistant_message: bool = False,
        **kwargs,
    ) -> None:
        """Initializes a new instance of the OumiDataset class."""
        if not hf_dataset_path:
            raise ValueError("The `hf_dataset_path` parameter must be provided.")
        if not messages_column:
            raise ValueError("The `messages_column` parameter must be provided.")
        self.messages_column = messages_column
        self.exclude_final_assistant_message = exclude_final_assistant_message
        kwargs["dataset_name"] = hf_dataset_path
        super().__init__(**kwargs)

    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Preprocesses the inputs of the example and returns a dictionary.

        Args:
            example: An example containing `messages` entries.

        Returns:
            Conversation: A Conversation object containing the messages.
        """
        messages = []

        if self.messages_column not in example:
            raise ValueError(
                f"The column '{self.messages_column}' is not present in the example."
            )
        example_messages = example[self.messages_column]

        for message in example_messages:
            if "role" not in message or "content" not in message:
                raise ValueError(
                    "The message format is invalid. Expected keys: 'role', 'content'."
                )
            if message["role"] == "user":
                role = Role.USER
            elif message["role"] == "assistant":
                role = Role.ASSISTANT
            else:
                raise ValueError(
                    f"Invalid role '{message['role']}'. Expected 'user' or 'assistant'."
                )
            content = message["content"] or ""
            messages.append(Message(role=role, content=content))

        if self.exclude_final_assistant_message and messages[-1].role == Role.ASSISTANT:
            messages = messages[:-1]

        return Conversation(messages=messages)



import math
import re

import pytest

from oumi.builders import build_reward_functions
from oumi.core.configs import TrainingParams
from oumi.core.registry import REGISTRY, RegistryType, register


@register("my_reward_func_starts_with_tldr", RegistryType.REWARD_FUNCTION)
def _starts_with_tldr_reward_func(completions, **kwargs):
    matches = [
        (content.startswith("TLDR") or content.startswith("TL;DR"))
        for content in completions
    ]
    return [1.0 if match else 0.0 for match in matches]


@register("my_reward_func_brevity", RegistryType.REWARD_FUNCTION)
def _brevity_func(completions, **kwargs):
    def _compute_reward(num_tokens, target_tokens=20):
        """Returns maximum reward for inputs that are `target_tokens` long"""
        x = float(num_tokens) / target_tokens
        return x * math.exp(-x)

    return [
        _compute_reward(len(re.split(r"\s+", content)), 20) for content in completions
    ]


def test_build_reward_functions_empty():
    assert len(build_reward_functions(TrainingParams())) == 0


@pytest.mark.parametrize(
    "function_name", ["my_reward_func_starts_with_tldr", "my_reward_func_brevity"]
)
def test_build_reward_functions_single(function_name: str):
    params = TrainingParams()
    params.reward_functions = [function_name]
    reward_funcs = build_reward_functions(params)
    assert len(reward_funcs) == 1
    assert reward_funcs == [REGISTRY.get(function_name, RegistryType.REWARD_FUNCTION)]

    params = TrainingParams()
    params.reward_functions = ["", function_name, ""]
    reward_funcs = build_reward_functions(params)
    assert len(reward_funcs) == 1
    assert reward_funcs == [REGISTRY.get(function_name, RegistryType.REWARD_FUNCTION)]


def test_build_reward_functions_multiple():
    params = TrainingParams()
    params.reward_functions = [
        "my_reward_func_starts_with_tldr",
        "my_reward_func_brevity",
    ]
    reward_funcs = build_reward_functions(params)
    assert len(reward_funcs) == 2
    assert reward_funcs == [
        REGISTRY.get("my_reward_func_starts_with_tldr", RegistryType.REWARD_FUNCTION),
        REGISTRY.get("my_reward_func_brevity", RegistryType.REWARD_FUNCTION),
    ]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union, cast

import datasets
import pandas as pd
from typing_extensions import override

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("nvidia/ChatQA-Training-Data")
class ChatqaDataset(BaseSftDataset):
    default_dataset = "nvidia/ChatQA-Training-Data"
    default_subset = "sft"

    def _get_system_message(self) -> Optional[str]:
        if self.dataset_subset == "sft":
            return None

        if self.dataset_subset == "synthetic_convqa":
            return "Please give a full and complete answer for the question."

        if self.dataset_subset in ("tatqa-arithmetic", "tatqa"):
            # Note: `tatqa` is a combination of `tatqa-arithmetic` and `tatqa-others``
            # Here we use the same prompt as tqta-arithmetic for both as it's larger
            # (8k vs 3k) Samples.
            # Preferably, each dataset should be loaded separately instead of loading
            # the combined `tatqa` subset.
            return (
                "Answer the following question with a number "
                "from context or the math arithmetic"
            )

        if self.dataset_subset == "tatqa-others":
            return (
                "Answer the following question with a short span, "
                "or a full and complete answer"
            )

        if self.dataset_subset in (
            "drop",
            "narrativeqa",
            "quoref",
            "ropes",
            "squad1.1",
            "squad2.0",
            "newsqa",
        ):
            return "Answer the following question with a short span."

        raise ValueError(f"Unknown dataset subset: {self.dataset_subset}")

    @override
    def transform_conversation(
        self, raw_conversation: Union[dict, pd.Series]
    ) -> Conversation:
        """Preprocesses the inputs of the example and returns a dictionary.

        ChatQA is a conversational question answering dataset.
        It contains 10 subsets. Some subsets contain grounding documents.

        See the dataset page for more information:
        https://huggingface.co/datasets/nvidia/ChatQA-Training-Data

        Args:
            raw_conversation: The raw conversation example.

        Returns:
            dict: The preprocessed inputs as an Oumi conversation.
        """
        messages = []

        # Step 1. Add system message. Most subsets contain one.
        system_message = self._get_system_message()
        if system_message:
            messages.append(Message(role=Role.SYSTEM, content=system_message))

        # Step 2. Add grounding context and system instruction
        has_context = raw_conversation.get("document") is not None
        if has_context:
            # Step 2.1. If the sample has a context, we add a system prompt
            # to only use information from the context to answer the question
            context_message = (
                "Only use the information from the user "
                "provided context to answer the question."
            )
            messages.append(Message(role=Role.SYSTEM, content=context_message))

            # Step 2.2. Add context document, wrapped in <context> tags
            # Note: This is not part of the original dataset
            # but is added to make the context more explicit.
            document = f"<context>{raw_conversation['document']}</document>"
            messages.append(Message(role=Role.USER, content=document))

        # Step 3. Add conversation history
        # Can contain one or multiple user/assistant turns.
        for message in raw_conversation["messages"]:
            messages.append(Message(role=message["role"], content=message["content"]))

        # Step 4. Add final assistant response, which is encoded differently
        # depending on the subset.
        if self.dataset_subset == "narrativeqa":
            # `narrativeqa` contains an array of arrays of strings
            # Note: All rows contain two answers.
            # We arbitrarily use the first one answer.
            response = raw_conversation["answers"][0][0]
        elif self.dataset_subset in ("squad1.1", "squad2.0"):
            # `squad1.1` and `squad2.0` contain a list of dicts
            # All rows contain a single answer.
            response = raw_conversation["answers"][0]["text"]
        else:
            # All other subsets contain a list of strings
            # All rows contain a single answer.
            response = raw_conversation["answers"][0]
        messages.append({"role": Role.ASSISTANT, "content": response})

        return Conversation(messages=messages)


@register_dataset("nvidia/ChatQA-Training-Data", subset="tatqa-arithmetic")
@register_dataset("nvidia/ChatQA-Training-Data", subset="tatqa-others")
class ChatqaTatqaDataset(ChatqaDataset):
    """ChatQA Subclass to handle tatqa subsets.

    The tatqa subsets require loading a specific file from
    the dataset repository, thus requiring us to override the
    default loading behavior.
    """

    default_subset = "tatqa-arithmetic"

    @override
    def _load_hf_hub_dataset(self) -> pd.DataFrame:
        if self.dataset_subset == "tatqa-arithmetic":
            filename = "tatqa/train_arithmetic.json"
        else:
            filename = "tatqa/train_others.json"

        if self.split is not None and self.split != "train":
            raise ValueError("Only the `train` split is supported for this dataset.")

        dataset = datasets.load_dataset(
            self.dataset_name, data_files={"train": filename}
        )
        dataset = cast(datasets.DatasetDict, dataset)
        return cast(pd.DataFrame, dataset["train"].to_pandas())



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy
import math
import warnings
from typing import cast, get_args

import torch
from typing_extensions import override

from oumi.builders import build_tokenizer
from oumi.core.configs import GenerationParams, InferenceConfig, ModelParams
from oumi.core.inference import BaseInferenceEngine
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.utils.conversation_utils import create_list_of_message_json_dicts
from oumi.utils.logging import logger
from oumi.utils.model_caching import get_local_filepath_for_gguf
from oumi.utils.peft_utils import get_lora_rank

try:
    import vllm  # pyright: ignore[reportMissingImports]

    try:
        from vllm.config import ModelDType  # pyright: ignore[reportMissingImports]
    except ImportError:
        # For compatibility with newer vLLM versions
        ModelDType = str  # type: ignore
    from vllm.entrypoints.chat_utils import (  # pyright: ignore[reportMissingImports]
        ChatCompletionMessageParam,
    )
    from vllm.lora.request import LoRARequest  # pyright: ignore[reportMissingImports]
    from vllm.model_executor.layers.quantization import (  # pyright: ignore[reportMissingImports]
        QuantizationMethods,
    )
    from vllm.sampling_params import (  # pyright: ignore[reportMissingImports]
        GuidedDecodingParams as VLLMGuidedDecodingParams,
    )
    from vllm.sampling_params import (  # pyright: ignore[reportMissingImports]
        SamplingParams,
    )
except ModuleNotFoundError:
    vllm = None


class VLLMInferenceEngine(BaseInferenceEngine):
    """Engine for running vLLM inference locally."""

    def __init__(
        self,
        model_params: ModelParams,
        *,
        generation_params: GenerationParams | None = None,
        tensor_parallel_size: int = -1,
        quantization: str | None = None,
        enable_prefix_caching: bool = True,
        gpu_memory_utilization: float = 0.9,
        enforce_eager: bool = True,
        max_num_seqs: int | None = None,
    ):
        """Initializes the inference Engine.

        Args:
            model_params: The model parameters to use for inference.
            generation_params: The generation parameters to use for inference.
            tensor_parallel_size: The number of tensor parallel processes to use.
                If set to -1, we will use all the available GPUs.
            quantization: The quantization method to use for inference.
            enable_prefix_caching: Whether to enable prefix caching.
            gpu_memory_utilization: The fraction of available GPU memory the model's
                executor will use. It can range from 0 to 1. Defaults to 0.9, i.e.,
                (90%) memory utilization.
            enforce_eager: Whether to enforce eager execution. Defaults to True.
                If False, will use eager mode and CUDA graph in hybrid mode.
            max_num_seqs: Maximum number of sequences per iteration.
        """
        super().__init__(model_params=model_params, generation_params=generation_params)

        if not vllm:
            raise RuntimeError(
                "vLLM is not installed. "
                "Please install the GPU dependencies for this package."
            )

        if not (
            math.isfinite(gpu_memory_utilization)
            and gpu_memory_utilization > 0
            and gpu_memory_utilization <= 1.0
        ):
            raise ValueError(
                "GPU memory utilization must be within (0, 1]. Got "
                f"{gpu_memory_utilization}."
            )

        # Infer the `quantization` type from the model's kwargs.
        if model_params.model_kwargs:
            if not quantization:
                # Check if quantization is BitsAndBytes.
                bnb_quantization_kwargs = ["load_in_4bit", "load_in_8bit"]
                for key in bnb_quantization_kwargs:
                    if model_params.model_kwargs.get(key):
                        quantization = "bitsandbytes"
                        break
                # Check if quantization is MXFP4.
                if not quantization and model_params.model_kwargs.get(
                    "quantization_config"
                ):
                    quant_config = model_params.model_kwargs.get("quantization_config")
                    if (
                        isinstance(quant_config, dict)
                        and quant_config.get("quant_method") == "mxfp4"
                    ):
                        quantization = "mxfp4"
            if not quantization and model_params.model_kwargs.get("filename"):
                # Check if quantization is GGUF.
                gguf_filename = str(model_params.model_kwargs.get("filename"))
                if gguf_filename.lower().endswith(".gguf"):
                    quantization = "gguf"
                    if (
                        not model_params.tokenizer_name
                        or model_params.tokenizer_name == model_params.model_name
                    ):
                        raise ValueError(
                            "GGUF quantization with the VLLM engine requires that you "
                            "explicitly set the `tokenizer_name` in `model_params`."
                        )

        vllm_kwargs = {}

        # Set the proper VLLM keys for the quantization type.
        if quantization and quantization == "bitsandbytes":
            vllm_kwargs["load_format"] = "bitsandbytes"
            logger.info("VLLM engine loading a `bitsandbytes` quantized model.")
        elif quantization and quantization == "mxfp4":
            # logic may not be needed; to be cleaned up after the next vllm patch
            # version release if possible
            # For MXFP4, set quantization in vllm_kwargs and clear variable
            # to avoid passing it twice
            vllm_kwargs["quantization"] = "mxfp4"
            quantization = None  # Avoid double setting
            logger.info("VLLM engine loading a `MXFP4` quantized model.")
        elif quantization and quantization == "gguf":
            # Download the GGUF file from HuggingFace to a local cache.
            gguf_local_path = get_local_filepath_for_gguf(
                repo_id=model_params.model_name,
                filename=gguf_filename,
            )
            # Overwrite `model_name` with the locally cached GGUF model.
            model_params = copy.deepcopy(model_params)
            model_params.model_name = gguf_local_path
            logger.info("VLLM engine loading a `GGUF` quantized model.")

        if tensor_parallel_size <= 0:
            if torch.cuda.device_count() > 1:
                tensor_parallel_size = torch.cuda.device_count()
            else:
                tensor_parallel_size = 1

        self._lora_request = None
        if model_params.adapter_model:
            # ID should be unique for this adapter, but isn't enforced by vLLM.
            self._lora_request = LoRARequest(
                lora_name="oumi_lora_adapter",
                lora_int_id=1,
                lora_path=model_params.adapter_model,
            )
            logger.info(f"Loaded LoRA adapter: {model_params.adapter_model}")
            lora_rank = get_lora_rank(model_params.adapter_model)
            vllm_kwargs["max_lora_rank"] = lora_rank
            logger.info(f"Setting vLLM max LoRA rank to {lora_rank}")

        if max_num_seqs is not None:
            vllm_kwargs["max_num_seqs"] = max_num_seqs

        self._tokenizer = build_tokenizer(model_params)

        supported_quantization_methods = list(get_args(QuantizationMethods))
        if quantization and quantization not in supported_quantization_methods:
            raise ValueError(
                f"Unsupported quantization method: {quantization}. "
                f"Supported methods are: {supported_quantization_methods}."
            )

        final_vllm_kwargs = dict(
            model=model_params.model_name,
            tokenizer=model_params.tokenizer_name,
            trust_remote_code=model_params.trust_remote_code,
            dtype=cast(ModelDType, model_params.torch_dtype_str),  # pyright: ignore[reportInvalidTypeForm]
            # TODO: these params should be settable via config,
            # but they don't belong to model_params
            tensor_parallel_size=tensor_parallel_size,
            enable_prefix_caching=enable_prefix_caching,
            enable_lora=self._lora_request is not None,
            max_model_len=model_params.model_max_length,
            gpu_memory_utilization=gpu_memory_utilization,
            enforce_eager=enforce_eager,
            **vllm_kwargs,
        )

        # Only add quantization if not already in vllm_kwargs and not None
        if quantization is not None and "quantization" not in vllm_kwargs:
            final_vllm_kwargs["quantization"] = quantization

        self._llm = vllm.LLM(**final_vllm_kwargs)  # pyright: ignore[reportArgumentType]
        # Ensure the tokenizer is set properly
        self._llm.set_tokenizer(self._tokenizer)

    def _convert_conversation_to_vllm_input(
        self, conversation: Conversation
    ) -> list[ChatCompletionMessageParam]:
        """Converts a conversation to a list of vllm input messages.

        Args:
            conversation: The conversation to convert.

        Returns:
            List[ChatCompletionMessageParam]: A list of vllm input messages.
        """
        result: list[ChatCompletionMessageParam] = []
        for json_dict in create_list_of_message_json_dicts(
            conversation.messages, group_adjacent_same_role_turns=True
        ):
            for key in ("role", "content"):
                if key not in json_dict:
                    raise RuntimeError(f"The required field '{key}' is missing!")
            if not isinstance(json_dict["content"], (str, list)):
                raise RuntimeError(
                    "The 'content' field must be `str` or `list`. "
                    f"Actual: {type(json_dict['content'])}."
                )
            result.append({"role": json_dict["role"], "content": json_dict["content"]})
        return result

    def _infer(
        self,
        input: list[Conversation],
        inference_config: InferenceConfig | None = None,
    ) -> list[Conversation]:
        """Runs model inference on the provided input.

        Documentation: https://docs.vllm.ai/en/stable/dev/sampling_params.html

        Args:
            input: A list of conversations to run inference on.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        generation_params = (
            inference_config.generation
            if inference_config and inference_config.generation
            else self._generation_params
        )

        model_params = (
            inference_config.model
            if inference_config and inference_config.model
            else self._model_params
        )

        if generation_params.guided_decoding is not None:
            guided_decoding = VLLMGuidedDecodingParams.from_optional(
                json=generation_params.guided_decoding.json,
                regex=generation_params.guided_decoding.regex,
                choice=generation_params.guided_decoding.choice,
            )
        else:
            guided_decoding = None

        sampling_params = SamplingParams(
            n=1,
            max_tokens=generation_params.max_new_tokens,
            temperature=generation_params.temperature,
            top_p=generation_params.top_p,
            frequency_penalty=generation_params.frequency_penalty,
            presence_penalty=generation_params.presence_penalty,
            stop=generation_params.stop_strings,
            stop_token_ids=generation_params.stop_token_ids,
            min_p=generation_params.min_p,
            guided_decoding=guided_decoding,
            skip_special_tokens=generation_params.skip_special_tokens,
        )

        output_conversations = []
        vllm_conversations = []
        non_skipped_conversations = []
        for conversation in input:
            if not conversation.messages:
                logger.warning("Conversation must have at least one message.")
                continue
            vllm_input = self._convert_conversation_to_vllm_input(conversation)
            vllm_conversations.append(vllm_input)
            non_skipped_conversations.append(conversation)

        if len(vllm_conversations) == 0:
            return []

        enable_tqdm = len(vllm_conversations) >= 2

        # Note: vLLM performs continuous batching under the hood.
        # We pass all the conversations and let vLLM handle the rest.
        chat_responses = self._llm.chat(
            vllm_conversations,
            sampling_params=sampling_params,
            lora_request=self._lora_request,
            use_tqdm=enable_tqdm,
            chat_template=None,
            chat_template_content_format="auto",
            chat_template_kwargs=model_params.chat_template_kwargs,
        )

        for conversation, chat_response in zip(
            non_skipped_conversations, chat_responses
        ):
            new_messages = [
                Message(content=message.text, role=Role.ASSISTANT)
                for message in chat_response.outputs
                if len(chat_response.outputs) > 0
            ]
            messages = [
                *conversation.messages,
                *new_messages,
            ]
            new_conversation = Conversation(
                messages=messages,
                metadata=conversation.metadata,
                conversation_id=conversation.conversation_id,
            )
            self._save_conversation_to_scratch(
                new_conversation,
                inference_config.output_path if inference_config else None,
            )
            output_conversations.append(new_conversation)

        return output_conversations

    def infer_online(
        self,
        input: list[Conversation],
        inference_config: InferenceConfig | None = None,
    ) -> list[Conversation]:
        """Runs model inference online.

        Args:
            input: A list of conversations to run inference on.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        warnings.warn(
            "infer_online() will be private in the future. Use infer() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        results = self._infer_online(input, inference_config)
        if inference_config and inference_config.output_path:
            self._save_conversations(results, inference_config.output_path)
        return results

    def infer_from_file(
        self,
        input_filepath: str,
        inference_config: InferenceConfig | None = None,
    ) -> list[Conversation]:
        """Runs model inference on inputs in the provided file.

        This is a convenience method to prevent boilerplate from asserting the
        existence of input_filepath in the generation_params.

        Args:
            input_filepath: Path to the input file containing prompts for
                generation.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        warnings.warn(
            "infer_from_file() will be private in the future. Use infer() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        input = self._read_conversations(input_filepath)
        results = self._infer(input, inference_config)
        if inference_config and inference_config.output_path:
            self._save_conversations(results, inference_config.output_path)
        return results

    @override
    def _infer_online(
        self,
        input: list[Conversation],
        inference_config: InferenceConfig | None = None,
    ) -> list[Conversation]:
        """Runs model inference online.

        Args:
            input: A list of conversations to run inference on.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        return self._infer(input, inference_config)

    @override
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine."""
        return {
            "frequency_penalty",
            "guided_decoding",
            "max_new_tokens",
            "min_p",
            "presence_penalty",
            "skip_special_tokens",
            "stop_strings",
            "stop_token_ids",
            "temperature",
            "top_p",
        }



import pytest

from oumi.datasets.grpo.rewards import (
    compute_sharp_target_token_length_reward,
    compute_soft_target_token_length_reward,
)


def test_compute_soft_target_token_length_reward():
    assert compute_soft_target_token_length_reward(
        0, target_tokens=10
    ) == pytest.approx(0, 1e-5)

    assert compute_soft_target_token_length_reward(
        1, target_tokens=10
    ) == pytest.approx(0.24596031, 1e-5)

    assert compute_soft_target_token_length_reward(
        5, target_tokens=10
    ) == pytest.approx(0.824360635, 1e-5)

    assert compute_soft_target_token_length_reward(
        10, target_tokens=10
    ) == pytest.approx(1.0, 1e-5)

    assert compute_soft_target_token_length_reward(
        20, target_tokens=10
    ) == pytest.approx(0.7357588, 1e-5)

    assert compute_soft_target_token_length_reward(
        100, target_tokens=10
    ) == pytest.approx(0.001234098, 1e-5)

    assert compute_soft_target_token_length_reward(
        1000, target_tokens=10
    ) == pytest.approx(0, 1e-5)


def test_compute_sharp_target_token_length_reward():
    assert compute_sharp_target_token_length_reward(0, target_tokens=10) == -10
    assert compute_sharp_target_token_length_reward(1, target_tokens=10) == -9
    assert compute_sharp_target_token_length_reward(10, target_tokens=10) == 0
    assert compute_sharp_target_token_length_reward(20, target_tokens=19) == -1
    assert compute_sharp_target_token_length_reward(100, target_tokens=20) == -80



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import aiohttp

_NON_RETRIABLE_STATUS_CODES = {
    400,  # Bad Request
    401,  # Unauthorized
    403,  # Forbidden
    404,  # Not Found
    422,  # Unprocessable Entity
}


def is_non_retriable_status_code(status_code: int) -> bool:
    """Check if a status code is non-retriable."""
    return status_code in _NON_RETRIABLE_STATUS_CODES


async def get_failure_reason_from_response(
    response: aiohttp.ClientResponse,
) -> str:
    """Return a string describing the error from the provided response."""
    try:
        response_json = await response.json()
        if isinstance(response_json, list):
            response_json = response_json[0]
        error_msg = (
            response_json.get("error", {}).get("message")
            if response_json
            else f"HTTP {response.status}"
        )

    except Exception:
        error_msg = f"HTTP {response.status}"

    return error_msg



import pytest
from transformers import AutoTokenizer

from oumi.core.registry import REGISTRY


def is_known_dataset_issue(dataset_name: str, idx: int) -> bool:
    """Check if the issue at the given index is a known issue."""
    known_issues = {
        "mlabonne/orpo-dpo-mix-40k": [
            15438,  # identical chosen and rejected responses
            16135,  # empty rejected key
            16798,  # identical chosen and rejected responses
            2173,  # identical chosen and rejected responses
            32553,  # identical chosen and rejected responses
            32572,  # identical chosen and rejected responses
            32579,  # identical chosen and rejected responses
            32580,  # identical chosen and rejected responses
            32670,  # identical chosen and rejected responses
            32671,  # identical chosen and rejected responses
            32672,  # identical chosen and rejected responses
            33072,  # identical chosen and rejected responses
            33073,  # identical chosen and rejected responses
            33142,  # identical chosen and rejected responses
            33195,  # identical chosen and rejected responses
            33203,  # identical chosen and rejected responses
            33207,  # identical chosen and rejected responses
            33217,  # identical chosen and rejected responses
            33221,  # identical chosen and rejected responses
            33224,  # identical chosen and rejected responses
            33228,  # identical chosen and rejected responses
            33241,  # identical chosen and rejected responses
            33479,  # identical chosen and rejected responses
            33511,  # identical chosen and rejected responses
            33646,  # identical chosen and rejected responses
            33703,  # identical chosen and rejected responses
            33739,  # identical chosen and rejected responses
            33773,  # identical chosen and rejected responses
            33878,  # identical chosen and rejected responses
            33886,  # identical chosen and rejected responses
            34031,  # identical chosen and rejected responses
            34163,  # identical chosen and rejected responses
            34188,  # identical chosen and rejected responses
            34422,  # identical chosen and rejected responses
            34450,  # identical chosen and rejected responses
            34487,  # identical chosen and rejected responses
            34552,  # identical chosen and rejected responses
            34646,  # identical chosen and rejected responses
            3585,  # identical chosen and rejected responses
            3841,  # identical chosen and rejected responses
            4544,  # identical chosen and rejected responses
            5670,  # identical chosen and rejected responses
            5776,  # empty rejected key
            6210,  # identical chosen and rejected responses
        ],
    }
    return idx in known_issues.get(dataset_name, [])


@pytest.fixture(
    params=[
        "mlabonne/orpo-dpo-mix-40k",
    ]
)
def dataset_fixture(request):
    dataset_name = request.param
    dataset_class = REGISTRY.get_dataset(dataset_name)
    if dataset_class is None:
        pytest.fail(f"Dataset {dataset_name} not found in registry")
    tokenizer = AutoTokenizer.from_pretrained("gpt2")
    return dataset_name, dataset_class(
        dataset_name=dataset_name, split="train", tokenizer=tokenizer
    )


@pytest.mark.skip(
    reason="This test is very time consuming, and should be run manually."
)
def test_dataset_structure(dataset_fixture):
    dataset_name, dataset = dataset_fixture
    assert len(dataset) > 0, f"Dataset {dataset_name} is empty"

    # Iterate through all items in the dataset
    for idx in range(len(dataset)):
        item = dataset[idx]

        # Check that each item has the expected keys
        assert "prompt" in item, f"'prompt' not found in item at index {idx}"
        assert "chosen" in item, f"'chosen' not found in item at index {idx}"
        assert "rejected" in item, f"'rejected' not found in item at index {idx}"

        # Check that the values are strings and not empty
        assert isinstance(item["prompt"], str), (
            f"'prompt' is not a string at index {idx}"
        )
        assert len(item["prompt"]) > 0, f"'prompt' is empty at index {idx}"

        assert isinstance(item["chosen"], str), (
            f"'chosen' is not a string at index {idx}"
        )
        assert len(item["chosen"]) > 0, f"'chosen' is empty at index {idx}"

        assert isinstance(item["rejected"], str), (
            f"'rejected' is not a string at index {idx}"
        )
        if len(item["rejected"]) == 0:
            if not is_known_dataset_issue(dataset_name, idx):
                pytest.fail(f"'rejected' is empty at index {idx}")
        else:
            assert len(item["rejected"]) > 0, f"'rejected' is empty at index {idx}"

        # Check that chosen and rejected are different, accounting for known issues
        if item["chosen"] == item["rejected"]:
            if not is_known_dataset_issue(dataset_name, idx):
                pytest.fail(
                    f"'chosen' and 'rejected' are identical at index {idx}, "
                    f"and this is not a known issue."
                )
        else:
            assert item["chosen"] != item["rejected"], (
                f"'chosen' and 'rejected' are identical at index {idx}"
            )



from unittest.mock import Mock, call, patch

import pytest

from oumi.core.configs import JobConfig
from oumi.core.launcher import JobState, JobStatus
from oumi.core.registry import REGISTRY, RegistryType
from oumi.launcher.clients.slurm_client import SlurmClient
from oumi.launcher.clouds.frontier_cloud import FrontierCloud
from oumi.launcher.clusters.frontier_cluster import FrontierCluster


#
# Fixtures
#
@pytest.fixture
def mock_slurm_client():
    with patch("oumi.launcher.clouds.frontier_cloud.SlurmClient") as client:
        yield client


@pytest.fixture
def mock_frontier_cluster():
    with patch("oumi.launcher.clouds.frontier_cloud.FrontierCluster") as cluster:
        cluster.SupportedQueues = FrontierCluster.SupportedQueues
        yield cluster


#
# Tests
#
def test_frontier_cloud_up_cluster_extended(mock_slurm_client, mock_frontier_cluster):
    cloud = FrontierCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=FrontierCluster)
    mock_frontier_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="extended.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = JobConfig(user="user")
    job_status = cloud.up_cluster(job, "extended.user")
    mock_slurm_client.assert_called_once_with(
        "user", "frontier.olcf.ornl.gov", "extended.user"
    )
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_frontier_cloud_up_cluster_batch(mock_slurm_client, mock_frontier_cluster):
    cloud = FrontierCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=FrontierCluster)
    mock_frontier_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="batch.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = JobConfig(user="user")
    job_status = cloud.up_cluster(job, "batch.user")
    mock_slurm_client.assert_called_once_with(
        "user", "frontier.olcf.ornl.gov", "batch.user"
    )
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_frontier_cloud_up_cluster_fails_mismatched_user(
    mock_slurm_client, mock_frontier_cluster
):
    cloud = FrontierCloud()
    with pytest.raises(ValueError, match="User must match the provided job user"):
        _ = cloud.up_cluster(JobConfig(user="user"), "batch.user1")
    with pytest.raises(ValueError, match="User must match the provided job user"):
        _ = cloud.up_cluster(JobConfig(user="user"), "extended.user1")


def test_frontier_cloud_up_cluster_default_queue(
    mock_slurm_client, mock_frontier_cluster
):
    cloud = FrontierCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=FrontierCluster)
    mock_frontier_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="batch.user",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = JobConfig(user="user")
    job_status = cloud.up_cluster(job, None)
    mock_slurm_client.assert_called_once_with(
        "user", "frontier.olcf.ornl.gov", "batch.user"
    )
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_frontier_cloud_init_with_users(mock_slurm_client):
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client, mock_client]
    mock_slurm_client.get_active_users.return_value = ["user1", "user2"]
    cloud = FrontierCloud()
    cluster_names = [cluster.name() for cluster in cloud.list_clusters()]
    cluster_names.sort()
    assert cluster_names == [
        "batch.user1",
        "batch.user2",
        "extended.user1",
        "extended.user2",
    ]
    mock_slurm_client.assert_has_calls(
        [
            call("user1", "frontier.olcf.ornl.gov", "batch.user1"),
            call("user2", "frontier.olcf.ornl.gov", "batch.user2"),
        ]
    )


def test_frontier_cloud_initialize_cluster(mock_slurm_client):
    cloud = FrontierCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client]
    clusters = cloud.initialize_clusters("me")
    clusters2 = cloud.initialize_clusters("me")
    mock_slurm_client.assert_called_once_with(
        "me", "frontier.olcf.ornl.gov", "batch.me"
    )
    cluster_names = [cluster.name() for cluster in clusters]
    cluster_names.sort()
    assert cluster_names == [
        "batch.me",
        "extended.me",
    ]
    # Verify that the second initialization returns the same clusters.
    assert clusters == clusters2


def test_frontier_cloud_list_clusters(mock_slurm_client):
    cloud = FrontierCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client, mock_client]
    # Check that there are no initial clusters.
    assert [] == cloud.list_clusters()
    cloud.initialize_clusters("me")
    clusters = cloud.list_clusters()
    expected_clusters = [
        "batch.me",
        "extended.me",
    ]
    cluster_names = [cluster.name() for cluster in clusters]
    cluster_names.sort()
    assert cluster_names == expected_clusters


def test_frontier_cloud_get_cluster_empty(mock_slurm_client):
    cloud = FrontierCloud()
    # Check that there are no initial clusters.
    assert cloud.get_cluster("batch.user") is None


def test_frontier_cloud_get_cluster_success(mock_slurm_client):
    cloud = FrontierCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client, mock_client]
    cloud.initialize_clusters("me")
    expected_clusters = [
        "batch.me",
        "extended.me",
    ]
    for name in expected_clusters:
        cluster = cloud.get_cluster(name)
        assert cluster is not None
        assert cluster.name() == name


def test_frontier_cloud_get_cluster_fails(mock_slurm_client):
    cloud = FrontierCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client, mock_client]
    cloud.initialize_clusters("me")
    assert cloud.get_cluster("nonexistent") is None


def test_frontier_cloud_builder_registered():
    assert REGISTRY.contains("frontier", RegistryType.CLOUD)



import tempfile
from pathlib import Path

import jsonlines
import pytest

from oumi.core.configs import GenerationParams, InferenceConfig, ModelParams
from oumi.core.types.conversation import (
    ContentItem,
    Conversation,
    Message,
    Role,
    Type,
)
from oumi.inference import NativeTextInferenceEngine
from oumi.utils.image_utils import load_image_png_bytes_from_path
from tests.integration.infer import get_default_device_map_for_inference
from tests.markers import requires_cuda_initialized, requires_gpus


def _get_default_text_model_params() -> ModelParams:
    return ModelParams(
        model_name="openai-community/gpt2",
        trust_remote_code=True,
        chat_template="gpt2",
        tokenizer_pad_token="<|endoftext|>",
        device_map=get_default_device_map_for_inference(),
    )


def _get_default_image_model_params() -> ModelParams:
    return ModelParams(
        model_name="Qwen/Qwen2-VL-2B-Instruct",
        model_max_length=1024,
        trust_remote_code=True,
        chat_template="qwen2-vl-instruct",
        device_map=get_default_device_map_for_inference(),
    )


def _get_default_inference_config() -> InferenceConfig:
    return InferenceConfig(
        generation=GenerationParams(
            max_new_tokens=5, use_sampling=False, temperature=0.0, min_p=0.0, seed=42
        )
    )


def _setup_input_conversations(filepath: str, conversations: list[Conversation]):
    Path(filepath).parent.mkdir(parents=True, exist_ok=True)
    Path(filepath).touch()
    with jsonlines.open(filepath, mode="w") as writer:
        for conversation in conversations:
            json_obj = conversation.to_dict()
            writer.write(json_obj)
    # Add some empty lines into the file
    with open(filepath, "a") as f:
        f.write("\n\n\n")


#
# Tests
#
@requires_gpus()
def test_infer_online():
    engine = NativeTextInferenceEngine(_get_default_text_model_params())
    conversation = Conversation(
        messages=[
            Message(
                content="Hello world!",
                role=Role.USER,
            ),
            Message(
                content="Hello again!",
                role=Role.USER,
            ),
        ],
        metadata={"foo": "bar"},
        conversation_id="123",
    )
    expected_result = [
        Conversation(
            messages=[
                *conversation.messages,
                Message(
                    content="The first time I saw",
                    role=Role.ASSISTANT,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
    ]
    result = engine.infer([conversation], _get_default_inference_config())
    assert expected_result == result


def test_infer_online_empty():
    engine = NativeTextInferenceEngine(_get_default_text_model_params())
    result = engine.infer([], _get_default_inference_config())
    assert [] == result


def test_infer_online_to_file():
    with tempfile.TemporaryDirectory() as output_temp_dir:
        engine = NativeTextInferenceEngine(_get_default_text_model_params())
        conversation_1 = Conversation(
            messages=[
                Message(
                    content="Hello world!",
                    role=Role.USER,
                ),
                Message(
                    content="Hello again!",
                    role=Role.USER,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
        conversation_2 = Conversation(
            messages=[
                Message(
                    content="Touche!",
                    role=Role.USER,
                ),
            ],
            metadata={"umi": "bar"},
            conversation_id="133",
        )
        expected_result = [
            Conversation(
                messages=[
                    *conversation_1.messages,
                    Message(
                        content="The first time I saw",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"foo": "bar"},
                conversation_id="123",
            ),
            Conversation(
                messages=[
                    *conversation_2.messages,
                    Message(
                        content="The U.S.",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"umi": "bar"},
                conversation_id="133",
            ),
        ]

        output_path = Path(output_temp_dir) / "b" / "output.jsonl"
        inference_config = _get_default_inference_config()
        inference_config.output_path = str(output_path)
        result = engine.infer(
            [conversation_1, conversation_2],
            inference_config,
        )
        assert result == expected_result
        with open(output_path) as f:
            parsed_conversations = []
            for line in f:
                parsed_conversations.append(Conversation.from_json(line))
            assert expected_result == parsed_conversations


def test_infer_from_file():
    with tempfile.TemporaryDirectory() as output_temp_dir:
        engine = NativeTextInferenceEngine(_get_default_text_model_params())
        conversation = Conversation(
            messages=[
                Message(
                    content="Hello world!",
                    role=Role.USER,
                ),
                Message(
                    content="Hello again!",
                    role=Role.USER,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
        input_path = Path(output_temp_dir) / "foo" / "input.jsonl"
        _setup_input_conversations(str(input_path), [conversation])
        expected_result = [
            Conversation(
                messages=[
                    *conversation.messages,
                    Message(
                        content="The first time I saw",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"foo": "bar"},
                conversation_id="123",
            )
        ]
        config = _get_default_inference_config()
        config.input_path = str(input_path)
        result = engine.infer(inference_config=config)
        assert expected_result == result


def test_infer_from_file_empty():
    with tempfile.TemporaryDirectory() as output_temp_dir:
        input_path = Path(output_temp_dir) / "foo" / "input.jsonl"
        _setup_input_conversations(str(input_path), [])
        engine = NativeTextInferenceEngine(_get_default_text_model_params())
        inference_config = _get_default_inference_config()
        inference_config.input_path = str(input_path)
        result = engine.infer(inference_config=inference_config)
        assert [] == result


def test_infer_from_file_to_file():
    with tempfile.TemporaryDirectory() as output_temp_dir:
        engine = NativeTextInferenceEngine(_get_default_text_model_params())
        conversation_1 = Conversation(
            messages=[
                Message(
                    content="Hello world!",
                    role=Role.USER,
                ),
                Message(
                    content="Hello again!",
                    role=Role.USER,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
        conversation_2 = Conversation(
            messages=[
                Message(
                    content="Touche!",
                    role=Role.USER,
                ),
            ],
            metadata={"umi": "bar"},
            conversation_id="133",
        )
        input_path = Path(output_temp_dir) / "foo" / "input.jsonl"
        _setup_input_conversations(str(input_path), [conversation_1, conversation_2])
        expected_result = [
            Conversation(
                messages=[
                    *conversation_1.messages,
                    Message(
                        content="The first time I saw",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"foo": "bar"},
                conversation_id="123",
            ),
            Conversation(
                messages=[
                    *conversation_2.messages,
                    Message(
                        content="The U.S.",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"umi": "bar"},
                conversation_id="133",
            ),
        ]

        output_path = Path(output_temp_dir) / "b" / "output.jsonl"
        inference_config = _get_default_inference_config()
        inference_config.output_path = str(output_path)
        result = engine.infer(
            [conversation_1, conversation_2],
            inference_config,
        )
        assert result == expected_result
        with open(output_path) as f:
            parsed_conversations = []
            for line in f:
                parsed_conversations.append(Conversation.from_json(line))
            assert expected_result == parsed_conversations


@requires_cuda_initialized()
@pytest.mark.single_gpu
def test_infer_from_file_to_file_with_images(root_testdata_dir: Path):
    png_image_bytes_great_wave = load_image_png_bytes_from_path(
        root_testdata_dir / "images" / "the_great_wave_off_kanagawa.jpg"
    )
    png_image_bytes_logo = load_image_png_bytes_from_path(
        root_testdata_dir / "images" / "oumi_logo_dark.png"
    )

    test_prompt: str = "Generate a short, descriptive caption for this image!"

    with tempfile.TemporaryDirectory() as output_temp_dir:
        engine = NativeTextInferenceEngine(_get_default_image_model_params())
        conversation_1 = Conversation(
            messages=[
                Message(
                    role=Role.USER,
                    content=[
                        ContentItem(
                            type=Type.IMAGE_BINARY,
                            binary=png_image_bytes_great_wave,
                        ),
                        ContentItem(
                            type=Type.TEXT,
                            content=test_prompt,
                        ),
                    ],
                )
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
        conversation_2 = Conversation(
            messages=[
                Message(
                    role=Role.USER,
                    content=[
                        ContentItem(
                            type=Type.IMAGE_BINARY,
                            binary=png_image_bytes_logo,
                        ),
                        ContentItem(
                            type=Type.TEXT,
                            content=test_prompt,
                        ),
                    ],
                ),
            ],
            metadata={"umi": "bar"},
            conversation_id="133",
        )
        input_path = Path(output_temp_dir) / "foo" / "input.jsonl"
        _setup_input_conversations(str(input_path), [conversation_1, conversation_2])

        output_path = Path(output_temp_dir) / "b" / "output.jsonl"
        inference_config = _get_default_inference_config()
        inference_config.output_path = str(output_path)

        result = engine.infer(
            [conversation_1, conversation_2],
            inference_config,
        )
        with open(output_path) as f:
            parsed_conversations = []
            for line in f:
                parsed_conversations.append(Conversation.from_json(line))
            assert result == parsed_conversations

        expected_results = [
            Conversation(
                messages=[
                    *conversation_1.messages,
                    Message(
                        content="",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"foo": "bar"},
                conversation_id="123",
            ),
            Conversation(
                messages=[
                    *conversation_2.messages,
                    Message(
                        content="",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"umi": "bar"},
                conversation_id="133",
            ),
        ]
        # Verify that the model response isn't empty, and verify that the results
        # are as expected except for the response content.
        assert len(result) == len(expected_results)
        for expected, actual in zip(expected_results, result):
            assert actual.messages[-1].content
            actual_dict = actual.to_dict()
            actual_dict["messages"][-1]["content"] = ""
            actual = Conversation.from_dict(actual_dict)
            assert actual == expected


def test_unsupported_model_raises_error():
    model_params = ModelParams(
        model_name="MlpEncoder",
        tokenizer_name="gpt2",
        tokenizer_pad_token="<|endoftext|>",
        load_pretrained_weights=False,
    )
    with pytest.raises(ValueError, match="requires a generation config"):
        NativeTextInferenceEngine(model_params)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from oumi.core.configs.analyze_config import (
    AnalyzeConfig,
    DatasetSource,
    SampleAnalyzerParams,
)


def test_dataset_source_required_field():
    """Test that dataset_source is a required field."""
    # Should work without dataset_source (MISSING becomes '???')
    config = AnalyzeConfig()
    assert config.dataset_source == "???"
    assert config.dataset_name == "Custom Dataset"

    # Should work with dataset_source
    config = AnalyzeConfig(
        dataset_source=DatasetSource.CONFIG, dataset_name="test_dataset"
    )
    assert config.dataset_source == DatasetSource.CONFIG
    assert config.dataset_name == "test_dataset"


def test_dataset_source_validation_success():
    """Test successful validation of dataset_source values."""
    # Test CONFIG mode
    config_config = AnalyzeConfig(
        dataset_source=DatasetSource.CONFIG, dataset_name="test_dataset"
    )
    assert config_config.dataset_source == DatasetSource.CONFIG
    assert config_config.dataset_name == "test_dataset"

    # Test DIRECT mode
    config_direct = AnalyzeConfig(
        dataset_source=DatasetSource.DIRECT, dataset_name="test_dataset"
    )
    assert config_direct.dataset_source == DatasetSource.DIRECT
    assert config_direct.dataset_name == "test_dataset"


def test_dataset_source_validation_invalid_value():
    """Test validation failure with invalid dataset_source value."""
    # The validation for invalid dataset_source values happens in DatasetAnalyzer,
    # not in AnalyzeConfig
    # So this should actually work (though it's not a valid enum value)
    config = AnalyzeConfig(dataset_source="invalid_value", dataset_name="test_dataset")  # type: ignore
    assert config.dataset_source == "invalid_value"

    # The validation will fail when trying to use this config in DatasetAnalyzer
    from oumi.core.analyze.dataset_analyzer import DatasetAnalyzer

    with pytest.raises(ValueError, match="Invalid dataset_source: invalid_value"):
        DatasetAnalyzer(config)


def test_sample_analyzer_param_validation_success():
    """Test successful validation of SampleAnalyzerParams."""
    # Should not raise any exception during __post_init__
    analyzer = SampleAnalyzerParams(id="test_analyzer")
    assert analyzer.id == "test_analyzer"


def test_sample_analyzer_param_validation_missing_id():
    """Test validation failure when id is missing."""
    with pytest.raises(ValueError, match="Analyzer 'id' must be provided"):
        AnalyzeConfig(
            dataset_source=DatasetSource.CONFIG,  # Required field
            dataset_name="test_dataset",
            analyzers=[SampleAnalyzerParams(id="")],
        )


def test_sample_analyzer_param_with_language_detection_params():
    """Test SampleAnalyzerParams with language detection analyzer configuration."""
    language_detection_params = {
        "confidence_threshold": 0.2,
        "top_k": 3,
        "multilingual_flag": {
            "enabled": True,
            "min_num_languages": 2,
        },
    }

    analyzer = SampleAnalyzerParams(
        id="language_detection", params=language_detection_params
    )

    assert analyzer.id == "language_detection"
    assert analyzer.params == language_detection_params


def test_analyze_config_validation_missing_dataset_name():
    """Test validation failure when dataset_name is missing."""
    with pytest.raises(
        ValueError,
        match="Either 'dataset_name' or 'dataset_path' must be provided when "
        "dataset_source=DatasetSource.CONFIG",
    ):
        AnalyzeConfig(
            dataset_source=DatasetSource.CONFIG,  # Required field
            dataset_name=None,
        )


def test_analyze_config_validation_empty_dataset_name():
    """Test validation failure when dataset_name is empty."""
    with pytest.raises(
        ValueError, match="Either 'dataset_name' or 'dataset_path' must be provided"
    ):
        AnalyzeConfig(
            dataset_source=DatasetSource.CONFIG,  # Required field
            dataset_name="",
        )


def test_analyze_config_validation_missing_dataset_path():
    """Test validation failure when both dataset_name and dataset_path are missing."""
    with pytest.raises(
        ValueError, match="Either 'dataset_name' or 'dataset_path' must be provided"
    ):
        AnalyzeConfig(
            dataset_source=DatasetSource.CONFIG,  # Required field
            dataset_name=None,
            dataset_path=None,
        )


def test_analyze_config_validation_empty_dataset_path():
    """Test validation failure when both dataset_name and dataset_path are empty."""
    with pytest.raises(
        ValueError, match="Either 'dataset_name' or 'dataset_path' must be provided"
    ):
        AnalyzeConfig(
            dataset_source=DatasetSource.CONFIG,  # Required field
            dataset_name="",
            dataset_path="",
        )


def test_analyze_config_validation_missing_processor_when_multimodal():
    """Test validation failure when processor_name is missing but
    is_multimodal is True."""
    with pytest.raises(
        ValueError,
        match="'processor_name' must be specified when 'is_multimodal' is True",
    ):
        AnalyzeConfig(
            dataset_source=DatasetSource.CONFIG,  # Required field
            dataset_path="/path/to/dataset.json",
            dataset_format="oumi",
            is_multimodal=True,
            processor_name=None,
        )


def test_analyze_config_validation_empty_processor_when_multimodal():
    """Test validation failure when processor_name is empty but is_multimodal
    is True."""
    with pytest.raises(
        ValueError,
        match="'processor_name' must be specified when 'is_multimodal' is True",
    ):
        AnalyzeConfig(
            dataset_path="/path/to/dataset.json",
            dataset_format="oumi",
            is_multimodal=True,
            processor_name="",
        )


def test_analyze_config_validation_multimodal_wrong_format():
    """Test validation failure when is_multimodal is True but dataset_format is
    not 'oumi'."""
    with pytest.raises(
        ValueError, match="Multimodal datasets require dataset_format='oumi'"
    ):
        AnalyzeConfig(
            dataset_path="/path/to/dataset.json",
            dataset_format="alpaca",
            is_multimodal=True,
            processor_name="openai/clip-vit-base-patch32",
        )


def test_analyze_config_validation_missing_dataset_format():
    """Test validation failure when dataset_path is provided but dataset_format
    is missing."""
    with pytest.raises(
        ValueError, match="'dataset_format' must be specified when using 'dataset_path'"
    ):
        AnalyzeConfig(dataset_path="/path/to/dataset.json")


def test_analyze_config_validation_invalid_dataset_format():
    """Test validation failure when dataset_format is not 'oumi' or 'alpaca'."""
    with pytest.raises(
        ValueError, match="'dataset_format' must be either 'oumi' or 'alpaca'"
    ):
        AnalyzeConfig(
            dataset_path="/path/to/dataset.json",
            dataset_format="invalid_format",
        )


def test_analyze_config_validation_missing_is_multimodal():
    """Test validation failure when dataset_path is provided but is_multimodal is
    missing."""
    with pytest.raises(
        ValueError, match="'is_multimodal' must be specified when using 'dataset_path'"
    ):
        AnalyzeConfig(
            dataset_path="/path/to/dataset.json",
            dataset_format="oumi",
            # Missing is_multimodal
        )


def test_analyze_config_validation_is_multimodal_required():
    """Test that is_multimodal can be explicitly set to True or False for
    custom datasets."""
    # Should work with is_multimodal=True
    config = AnalyzeConfig(
        dataset_path="/path/to/dataset.json",
        dataset_format="oumi",
        is_multimodal=True,
        processor_name="openai/clip-vit-base-patch32",
    )
    assert config.is_multimodal is True

    # Should work with is_multimodal=False
    config = AnalyzeConfig(
        dataset_path="/path/to/dataset.json",
        dataset_format="oumi",
        is_multimodal=False,
    )
    assert config.is_multimodal is False


def test_analyze_config_validation_with_valid_analyzers():
    """Test validation with valid analyzer configurations."""
    analyzers = [
        SampleAnalyzerParams(id="analyzer1"),
        SampleAnalyzerParams(id="analyzer2"),
    ]

    # Should not raise any exception during __post_init__
    AnalyzeConfig(dataset_name="test_dataset", analyzers=analyzers)


def test_analyze_config_validation_duplicate_analyzer_ids():
    """Test validation failure with duplicate analyzer IDs."""
    analyzers = [
        SampleAnalyzerParams(id="duplicate_id"),
        SampleAnalyzerParams(id="duplicate_id"),
    ]

    with pytest.raises(ValueError, match="Duplicate analyzer ID found: 'duplicate_id'"):
        AnalyzeConfig(dataset_name="test_dataset", analyzers=analyzers)


def test_analyze_config_default_values():
    """Test that AnalyzeConfig has correct default values."""
    config = AnalyzeConfig(
        dataset_source=DatasetSource.CONFIG, dataset_name="test_dataset"
    )

    assert config.dataset_source == DatasetSource.CONFIG
    assert config.dataset_name == "test_dataset"
    assert config.split == "train"  # default value
    assert config.sample_count is None  # default value
    assert config.output_path == "."  # default value
    assert config.analyzers == []  # default value


def test_analyze_config_with_custom_values():
    """Test AnalyzeConfig with custom parameter values."""
    analyzers = [
        SampleAnalyzerParams(id="analyzer1", params={"param1": "value1"}),
        SampleAnalyzerParams(id="analyzer2", params={"param2": "value2"}),
    ]

    config = AnalyzeConfig(
        dataset_name="test_dataset",
        split="test",
        sample_count=100,
        output_path="/tmp/output",
        analyzers=analyzers,
        processor_name="Salesforce/blip2-opt-2.7b",
        processor_kwargs={"image_size": 224, "do_resize": True},
        trust_remote_code=True,
    )

    assert config.dataset_name == "test_dataset"
    assert config.split == "test"
    assert config.sample_count == 100
    assert config.output_path == "/tmp/output"
    assert len(config.analyzers) == 2
    assert config.analyzers[0].id == "analyzer1"
    assert config.analyzers[1].id == "analyzer2"
    assert config.processor_name == "Salesforce/blip2-opt-2.7b"
    assert config.processor_kwargs == {"image_size": 224, "do_resize": True}
    assert config.trust_remote_code is True


def test_analyze_config_processor_fields_custom_values():
    """Test AnalyzeConfig with custom processor parameter values."""
    config = AnalyzeConfig(
        dataset_name="test_dataset",
        processor_name="Salesforce/blip2-opt-2.7b",
        processor_kwargs={"image_size": 224, "do_resize": True},
        trust_remote_code=True,
    )

    assert config.processor_name == "Salesforce/blip2-opt-2.7b"
    assert config.processor_kwargs == {"image_size": 224, "do_resize": True}
    assert config.trust_remote_code is True


def test_analyze_config_sample_count_zero():
    """Test validation failure when sample_count is zero."""
    with pytest.raises(ValueError, match="`sample_count` must be greater than 0."):
        AnalyzeConfig(dataset_name="test_dataset", sample_count=0)


def test_analyze_config_sample_count_negative():
    """Test validation failure when sample_count is negative."""
    with pytest.raises(ValueError, match="`sample_count` must be greater than 0."):
        AnalyzeConfig(dataset_name="test_dataset", sample_count=-5)


def test_analyze_config_sample_count_valid():
    """Test that valid sample_count values are accepted."""
    # Should not raise any exception
    config = AnalyzeConfig(dataset_name="test_dataset", sample_count=1)
    assert config.sample_count == 1

    config = AnalyzeConfig(dataset_name="test_dataset", sample_count=100)
    assert config.sample_count == 100

    # None should also be valid
    config = AnalyzeConfig(dataset_name="test_dataset", sample_count=None)
    assert config.sample_count is None



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("cerebras/SlimPajama-627B")
class SlimPajamaDataset(BasePretrainingDataset):
    """SlimPajama-627B: A cleaned and deduplicated version of RedPajama.

    SlimPajama is the largest extensively deduplicated, multi-corpora, open-source
    dataset for training large language models. It was created by cleaning and
    deduplicating the 1.2T token RedPajama dataset, resulting in a 627B token dataset.

    The dataset consists of 59166 jsonl files and is ~895GB compressed. It includes
    training, validation, and test splits :footcite:`2023_slimpajama`.

    Key Features:
        - 627B tokens
        - Open-source
        - Curated data sources
        - Extensive deduplication
        - Primarily English language

    Data Sources and Proportions:
        - Commoncrawl: 52.2%
        - C4: 26.7%
        - GitHub: 5.2%
        - Books: 4.2%
        - ArXiv: 4.6%
        - Wikipedia: 3.8%
        - StackExchange: 3.3%

    See Also:
        - Blog post: https://www.cerebras.net/blog/slimpajama-a-627b-token-cleaned-and-deduplicated-version-of-redpajama
        - Repository: https://github.com/Cerebras/modelzoo/tree/main/modelzoo/transformers/data_processing/slimpajama
        - Hugging Face: https://huggingface.co/datasets/cerebras/SlimPajama-627B

    Dataset Structure:
        Each example is a JSON object with the following structure:

        .. code-block:: python

            {
                "text": str,
                "meta": {
                    "redpajama_set_name": str  # One of the data source names
                }
            }

    Citations:
        .. footbibliography::
    """

    default_dataset = "cerebras/SlimPajama-627B"



import os
import re
from typing import Optional
from unittest.mock import patch

import pytest

from oumi.builders import build_tokenizer
from oumi.core.configs import ModelParams
from oumi.utils.str_utils import (
    compute_utf8_len,
    get_editable_install_override_env_var,
    sanitize_run_name,
    set_oumi_install_editable,
    str_to_bool,
    truncate_text_pieces_to_max_tokens_limit,
    truncate_to_max_tokens_limit,
    try_str_to_bool,
)


@pytest.fixture
def gpt2_tokenizer():
    tokenizer = build_tokenizer(
        ModelParams(
            model_name="openai-community/gpt2",
            torch_dtype_str="float16",
            trust_remote_code=False,
            chat_template="default",
            tokenizer_pad_token="<|endoftext|>",
        )
    )
    assert tokenizer.pad_token_id is not None
    assert isinstance(tokenizer.pad_token_id, int)
    return tokenizer


def test_sanitize_run_name_empty():
    assert sanitize_run_name("") == ""


def test_sanitize_run_name_below_max_length_limit():
    assert sanitize_run_name("abc.XYZ-0129_") == "abc.XYZ-0129_"
    assert sanitize_run_name("a_X-7." * 16) == "a_X-7." * 16
    assert sanitize_run_name("a" * 99) == "a" * 99
    assert sanitize_run_name("X" * 100) == "X" * 100


def test_sanitize_run_name_below_invalid_chars():
    assert sanitize_run_name("abc?XYZ/0129^") == "abc_XYZ_0129_"
    assert sanitize_run_name("Лемма") == "_____"


def test_sanitize_run_name_too_long():
    raw_long_run_name = (
        "fineweb.pt.FSDP.HYBRID_SHARD.4node.4xA10040GB.20steps.bs16.gas16.v907."
        "sky-2024-07-22-16-26-33-541717_xrdaukar-4node4gpu-01-oumi-cluster"
    )
    actual = sanitize_run_name(raw_long_run_name)
    assert actual is not None
    assert len(actual) == 100
    expected = (
        "fineweb.pt.FSDP.HYBRID_SHARD.4node.4xA10040GB.20steps.bs16.gas16.v907."
        "sky-2024-07...9db6edce8186fb6a"
    )
    assert actual == expected
    # verify it's idempotent
    assert sanitize_run_name(actual) == expected


@pytest.mark.parametrize(
    "value",
    ["true", "True", "TRUE", "yes", "Yes", "YES", "1", "on", "ON", "t", "y", " True "],
)
def test_true_values(value):
    assert str_to_bool(value) is True
    assert try_str_to_bool(value) is True


@pytest.mark.parametrize(
    "value",
    [
        "false",
        "False",
        "FALSE",
        "no",
        "No",
        "NO",
        "0",
        "off",
        "OFF",
        "f",
        "n",
        " False ",
    ],
)
def test_false_values(value):
    assert str_to_bool(value) is False
    assert try_str_to_bool(value) is False


@pytest.mark.parametrize("value", ["maybe", "unknown", "tru", "ye", "2", "nope"])
def test_invalid_inputs(value):
    assert try_str_to_bool(value) is None
    with pytest.raises(ValueError):
        str_to_bool(value)


def test_compute_utf8_len():
    assert compute_utf8_len("") == 0
    assert compute_utf8_len("a") == 1
    assert compute_utf8_len("abc") == 3
    assert compute_utf8_len("a b c") == 5
    assert compute_utf8_len("Wir müssen") == 11
    assert compute_utf8_len("Мы должны") == 17


@pytest.mark.parametrize(
    "env_var_val,expected_val",
    [
        (None, False),
        ("1", True),
        ("true", True),
        ("True", True),
        ("0", False),
        ("false", False),
        ("False", False),
    ],
)
def test_get_editable_install_override(env_var_val: Optional[str], expected_val: bool):
    overrides = {}
    if env_var_val is not None:
        overrides = {"OUMI_FORCE_EDITABLE_INSTALL": env_var_val}
    with patch.dict(os.environ, overrides, clear=True):
        assert get_editable_install_override_env_var() == expected_val


@pytest.mark.parametrize(
    "setup,output_setup",
    [
        (
            "pip install 'oumi[gpu]'",
            "pip install -e '.[gpu]'",
        ),
        (
            """
            #A comment
            pip install -e uv && uv pip -q install "oumi[gpu,dev]" vllm # comment
            pip install -e "oumi"
            """,
            """
            #A comment
            pip install -e uv && uv pip -q install -e '.[gpu,dev]' vllm # comment
            pip install -e "oumi"
            """,
        ),
        (
            """
            #A comment
            pip -q --debug install -U "skypilot[azure]" oumi vllm "wandb" # Foo.
            print("All done")
            """,
            """
            #A comment
            pip -q --debug install -U "skypilot[azure]" -e '.' vllm "wandb" # Foo.
            print("All done")
            """,
        ),
    ],
)
def test_set_oumi_install_editable(setup, output_setup):
    actual_setup = set_oumi_install_editable(setup)
    assert actual_setup == output_setup


@pytest.mark.parametrize(
    "text,max_tokens,truncation_side,expected_text,expected_tokens",
    [
        ("Hello World!", 100, "right", "Hello World!", 3),
        ("Hello World!", 100, "left", "Hello World!", 3),
        ("Hello World!", 3, "right", "Hello World!", 3),
        ("Hello World!", 3, "left", "Hello World!", 3),
        ("Hello World!", 2, "right", "Hello World", 2),
        ("Hello World!", 2, "left", " World!", 2),
        ("Hello World!", 1, "right", "Hello", 1),
        ("Hello World!", 1, "left", "!", 1),
        ("", 10, "right", "", 0),
        ("", 10, "left", "", 0),
        ("", 1, "right", "", 0),
        ("", 1, "left", "", 0),
        ("  \t", 10, "right", "  \t", 3),
        ("  \t\n", 10, "left", "  \t\n", 4),
        ("  \t", 1, "right", " ", 1),
        ("  \t\n", 1, "left", "\n", 1),
        ("Hellolololworld!", 1, "right", "Hell", 1),
        ("Hellolololworld!", 1, "left", "!", 1),
        ("Hellolololworld!", 2, "right", "Hellol", 2),
        ("Hellolololworld!", 2, "left", "world!", 2),
    ],
)
def test_truncate_to_max_tokens_limit_success(
    text: str,
    max_tokens: int,
    truncation_side: str,
    expected_text: str,
    expected_tokens: int,
    gpt2_tokenizer,
):
    truncated_text, truncated_tokens = truncate_to_max_tokens_limit(
        text,
        tokenizer=gpt2_tokenizer,
        max_tokens=max_tokens,
        truncation_side=truncation_side,
    )
    assert truncated_text == expected_text
    assert truncated_tokens == expected_tokens


def test_truncate_to_max_tokens_limit_invalid_args(
    gpt2_tokenizer,
):
    with pytest.raises(
        ValueError, match=re.escape("`max_tokens` must be a positive integer")
    ):
        truncated_text, truncated_tokens = truncate_to_max_tokens_limit(
            "hi there",
            tokenizer=gpt2_tokenizer,
            max_tokens=0,
            truncation_side="right",
        )

    with pytest.raises(
        ValueError, match=re.escape("`max_tokens` must be a positive integer")
    ):
        truncated_text, truncated_tokens = truncate_to_max_tokens_limit(
            "apple",
            tokenizer=gpt2_tokenizer,
            max_tokens=-1,
            truncation_side="left",
        )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Invalid truncation_side: 'kaboom'. Expected 'left' or 'right'"
        ),
    ):
        truncated_text, truncated_tokens = truncate_to_max_tokens_limit(
            "hi there",
            tokenizer=gpt2_tokenizer,
            max_tokens=100,
            truncation_side="kaboom",
        )


@pytest.mark.parametrize(
    "text_pieces,max_tokens,truncation_side,expected_text_pieces",
    [
        (["Hello", "World!"], 100, "right", None),
        (["Hello", "World!"], 100, "left", None),
        (["Hello", "World!"], 2, "right", ["Hello", "World"]),
        (["Hello", "World!"], 2, "left", ["", "World!"]),
        (["Hello", "World!"], 1, "right", ["Hello", ""]),
        (["Hello", "World!"], 1, "left", ["", "!"]),
    ],
)
def test_truncate_text_pieces_to_max_tokens_limit_success(
    text_pieces: list[str],
    max_tokens: int,
    truncation_side: str,
    expected_text_pieces: Optional[list[str]],
    gpt2_tokenizer,
):
    truncated_text_pieces = truncate_text_pieces_to_max_tokens_limit(
        text_pieces,
        tokenizer=gpt2_tokenizer,
        max_tokens=max_tokens,
        truncation_side=truncation_side,
    )

    if expected_text_pieces is not None:
        assert truncated_text_pieces == expected_text_pieces
    else:
        assert truncated_text_pieces == text_pieces



import contextlib
import inspect
from importlib.util import find_spec
from unittest import mock
from unittest.mock import patch

import pytest

from oumi.core.configs import (
    GenerationParams,
    InferenceConfig,
    ModelParams,
    RemoteParams,
)
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.inference import (
    AnthropicInferenceEngine,
    GoogleVertexInferenceEngine,
    LlamaCppInferenceEngine,
    NativeTextInferenceEngine,
    RemoteInferenceEngine,
    RemoteVLLMInferenceEngine,
    SambanovaInferenceEngine,
    SGLangInferenceEngine,
    VLLMInferenceEngine,
)

vllm_import_failed = find_spec("vllm") is None
llama_cpp_import_failed = find_spec("llama_cpp") is None

SUPPORTED_INFERENCE_ENGINES = [
    RemoteInferenceEngine,
    AnthropicInferenceEngine,
    LlamaCppInferenceEngine,
    NativeTextInferenceEngine,
    SambanovaInferenceEngine,
    SGLangInferenceEngine,
    VLLMInferenceEngine,
    RemoteVLLMInferenceEngine,
    GoogleVertexInferenceEngine,
]


@pytest.fixture
def sample_conversation():
    return Conversation(
        messages=[
            Message(role=Role.USER, content="Hello, how are you?"),
        ]
    )


@pytest.fixture
def model_params():
    return ModelParams(
        model_name="openai-community/gpt2",
        tokenizer_pad_token="<|endoftext|>",
        tokenizer_name="gpt2",
        load_pretrained_weights=False,
    )


@pytest.fixture
def generation_params_fields():
    """Get all field names from GenerationParams."""
    return set(inspect.signature(GenerationParams).parameters.keys())


def _should_skip_engine(engine_class) -> bool:
    return (engine_class == VLLMInferenceEngine and vllm_import_failed) or (
        engine_class == LlamaCppInferenceEngine and llama_cpp_import_failed
    )


def _mock_engine(engine_class):
    """Mock the engine to avoid loading non-existent models."""

    mock_tokenizer = mock.MagicMock()
    mock_tokenizer.pad_token_id = 0
    mock_tokenizer.eos_token_id = 0
    mock_tokenizer.eos_token = "<eos>"
    mock_tokenizer.batch_decode = mock.MagicMock()
    mock_tokenizer.batch_decode.return_value = ["I'm fine, how are you?"]
    mock_tokenizer.apply_chat_template = mock.MagicMock()
    mock_tokenizer.apply_chat_template.return_value = (
        "<|startoftext|>I'm fine, how are you? <|endoftext|>"
    )
    mock_model = mock.MagicMock()
    mock_model.generate = mock.MagicMock()  # Add generate attribute

    if engine_class == VLLMInferenceEngine:
        mock_llm = mock.MagicMock()
        mock_ctx = patch.multiple(
            "oumi.inference.vllm_inference_engine",
            vllm=mock.MagicMock(LLM=mock.MagicMock(return_value=mock_llm)),
            build_tokenizer=mock.MagicMock(return_value=mock_tokenizer),
        )
    elif engine_class == LlamaCppInferenceEngine:
        mock_ctx = patch("llama_cpp.Llama.from_pretrained")
    elif engine_class == SGLangInferenceEngine:
        mock_ctx = patch.multiple(
            "oumi.inference.sglang_inference_engine",
            build_tokenizer=mock.MagicMock(return_value=mock_tokenizer),
            build_processor=mock.MagicMock(return_value=None),
            is_image_text_llm=mock.MagicMock(return_value=False),
        )
    elif engine_class == NativeTextInferenceEngine:
        mock_ctx = patch.multiple(
            "oumi.inference.native_text_inference_engine",
            build_model=mock.MagicMock(return_value=mock_model),
            build_tokenizer=mock.MagicMock(return_value=mock_tokenizer),
            build_processor=mock.MagicMock(return_value=None),
            is_image_text_llm=mock.MagicMock(return_value=False),
        )
    elif issubclass(engine_class, RemoteInferenceEngine):
        mock_ctx = patch("aiohttp.ClientSession")
    else:
        mock_ctx = contextlib.nullcontext()

    return mock_ctx


def test_generation_params_validation():
    with pytest.raises(ValueError, match="Temperature must be non-negative."):
        GenerationParams(temperature=-0.1)

    with pytest.raises(ValueError, match="top_p must be between 0 and 1."):
        GenerationParams(top_p=1.1)

    with pytest.raises(
        ValueError, match="Logit bias for token 1 must be between -100 and 100."
    ):
        GenerationParams(logit_bias={1: 101})

    with pytest.raises(ValueError, match="min_p must be between 0 and 1."):
        GenerationParams(min_p=1.1)


@pytest.mark.parametrize(
    "engine_class",
    SUPPORTED_INFERENCE_ENGINES,
)
def test_generation_params_used_in_inference(
    engine_class, sample_conversation, model_params
):
    if _should_skip_engine(engine_class):
        pytest.skip(f"{engine_class.__name__} is not available")

    mock_ctx = _mock_engine(engine_class)

    with (
        patch.object(
            engine_class, "_infer", return_value=[sample_conversation]
        ) as mock_infer,
        mock_ctx,
    ):
        remote_params = RemoteParams(api_url="<placeholder>")
        if issubclass(engine_class, RemoteInferenceEngine):
            engine = engine_class(
                model_params=model_params, remote_params=remote_params
            )
        else:
            engine = engine_class(model_params)

        generation_params = GenerationParams(
            max_new_tokens=100,
            temperature=0.7,
            top_p=0.9,
            frequency_penalty=0.1,
            presence_penalty=0.1,
            stop_strings=["END"],
            stop_token_ids=[128001, 128008, 128009],
            logit_bias={1: 1.0, 2: -1.0},
            min_p=0.05,
        )
        inference_config = InferenceConfig(
            model=model_params,
            generation=generation_params,
            remote_params=remote_params,
        )

        result = engine.infer([sample_conversation], inference_config)

        # Check that the result is as expected
        assert result == [sample_conversation]

        # Check that _infer was called with the correct parameters
        mock_infer.assert_called_once()
        called_params = mock_infer.call_args[0][1].generation
        assert called_params.max_new_tokens == 100
        assert called_params.temperature == 0.7
        assert called_params.top_p == 0.9
        assert called_params.frequency_penalty == 0.1
        assert called_params.presence_penalty == 0.1
        assert called_params.stop_strings == ["END"]
        assert called_params.stop_token_ids == [128001, 128008, 128009]
        assert called_params.logit_bias == {1: 1.0, 2: -1.0}
        assert called_params.min_p == 0.05


@pytest.mark.parametrize(
    "engine_class",
    SUPPORTED_INFERENCE_ENGINES,
)
def test_generation_params_defaults_used_in_inference(
    engine_class, sample_conversation, model_params
):
    if _should_skip_engine(engine_class):
        pytest.skip(f"{engine_class.__name__} is not available")

    mock_ctx = _mock_engine(engine_class)

    with (
        patch.object(
            engine_class, "_infer", return_value=[sample_conversation]
        ) as mock_infer,
        mock_ctx,
    ):
        remote_params = RemoteParams(api_url="<placeholder>")
        if issubclass(engine_class, RemoteInferenceEngine):
            engine = engine_class(
                model_params=model_params, remote_params=remote_params
            )
        else:
            engine = engine_class(model_params)

        generation_params = GenerationParams()
        inference_config = InferenceConfig(
            model=model_params,
            generation=generation_params,
            remote_params=remote_params,
        )

        result = engine.infer([sample_conversation], inference_config)

        assert result == [sample_conversation]

        mock_infer.assert_called_once()
        called_params = mock_infer.call_args[0][1].generation
        assert called_params.max_new_tokens == 1024
        assert called_params.temperature == 0.0
        assert called_params.top_p == 1.0
        assert called_params.frequency_penalty == 0.0
        assert called_params.presence_penalty == 0.0
        assert called_params.stop_strings is None
        assert called_params.logit_bias == {}
        assert called_params.min_p == 0.0


@pytest.mark.parametrize(
    "engine_class",
    SUPPORTED_INFERENCE_ENGINES,
)
def test_supported_params_exist_in_config(
    engine_class, model_params, generation_params_fields
):
    if _should_skip_engine(engine_class):
        pytest.skip(f"{engine_class.__name__} is not available")

    mock_ctx = _mock_engine(engine_class)

    with mock_ctx:
        remote_params = RemoteParams(api_url="<placeholder>")
        if issubclass(engine_class, RemoteInferenceEngine):
            engine = engine_class(
                model_params=model_params, remote_params=remote_params
            )
        else:
            engine = engine_class(model_params)

        supported_params = engine.get_supported_params()

        # Additional check that all expected params exist in GenerationParams
        invalid_params = supported_params - generation_params_fields

        assert not invalid_params, (
            f"Test expects support for parameters that don't exist in "
            f"GenerationParams: {invalid_params}"
        )


@pytest.mark.parametrize(
    "engine_class,unsupported_param,value",
    [
        (AnthropicInferenceEngine, "min_p", 0.1),
        (AnthropicInferenceEngine, "frequency_penalty", 0.5),
        (GoogleVertexInferenceEngine, "frequency_penalty", 0.5),
        (GoogleVertexInferenceEngine, "presence_penalty", 0.5),
        (VLLMInferenceEngine, "logit_bias", {1: 1.0}),
        (LlamaCppInferenceEngine, "num_beams", 8),
    ],
)
def test_unsupported_params_warning(
    engine_class, unsupported_param, value, model_params, sample_conversation, caplog
):
    if _should_skip_engine(engine_class):
        pytest.skip(f"{engine_class.__name__} is not available")

    mock_ctx = _mock_engine(engine_class)

    with (
        mock_ctx,
        patch.object(engine_class, "_infer", return_value=[sample_conversation]),
    ):
        remote_params = RemoteParams(api_url="test")
        if issubclass(engine_class, RemoteInferenceEngine):
            engine = engine_class(
                model_params=model_params, remote_params=remote_params
            )
        else:
            engine = engine_class(model_params)

        # Create generation params with the unsupported parameter
        params_dict = {
            "max_new_tokens": 100,  # Add a supported param
            unsupported_param: value,
        }

        generation_params = GenerationParams(**params_dict)
        inference_config = InferenceConfig(
            model=model_params, generation=generation_params
        )
        if issubclass(engine_class, RemoteInferenceEngine):
            inference_config.remote_params = remote_params

        # Call infer which should trigger the warning
        engine.infer([sample_conversation], inference_config)

        # Check that warning was logged
        assert any(
            record.levelname == "WARNING"
            and f"{engine_class.__name__} does not support {unsupported_param}"
            in record.message
            for record in caplog.records
        )


@pytest.mark.parametrize(
    "engine_class,param,default_value",
    [
        (AnthropicInferenceEngine, "min_p", 0.0),
        (AnthropicInferenceEngine, "frequency_penalty", 0.0),
        (VLLMInferenceEngine, "logit_bias", {}),
        (LlamaCppInferenceEngine, "num_beams", 1),
    ],
)
def test_no_warning_for_default_values(
    engine_class, param, default_value, model_params, sample_conversation, caplog
):
    if _should_skip_engine(engine_class):
        pytest.skip(f"{engine_class.__name__} is not available")

    mock_ctx = _mock_engine(engine_class)

    with (
        mock_ctx,
        patch.object(engine_class, "_infer", return_value=[sample_conversation]),
    ):
        remote_params = RemoteParams(api_url="test")
        if issubclass(engine_class, RemoteInferenceEngine):
            engine = engine_class(
                model_params=model_params, remote_params=remote_params
            )
        else:
            engine = engine_class(model_params)

        params_dict = {
            "max_new_tokens": 100,  # Add a supported param
            param: default_value,
        }

        generation_params = GenerationParams(**params_dict)
        inference_config = InferenceConfig(
            model=model_params, generation=generation_params
        )
        if issubclass(engine_class, RemoteInferenceEngine):
            inference_config.remote_params = remote_params

        engine.infer([sample_conversation], inference_config)

        # Check that no warning was logged for this parameter
        assert not any(
            record.levelname == "WARNING"
            and f"{engine_class.__name__} does not support {param}" in record.message
            for record in caplog.records
        )


@pytest.mark.parametrize(
    "engine_class",
    SUPPORTED_INFERENCE_ENGINES,
)
def test_supported_params_are_accessed(engine_class, model_params, sample_conversation):
    """Test that all supported parameters are actually accessed during inference."""
    if _should_skip_engine(engine_class):
        pytest.skip(f"{engine_class.__name__} is not available")

    mock_ctx = _mock_engine(engine_class)

    class AccessTrackingGenerationParams(GenerationParams):
        """A version of GenerationParams that tracks which parameters are accessed."""

        _accessed_params: set[str]
        _track_access: bool

        def __init__(self, **kwargs):
            self._accessed_params: set[str] = set()
            self._track_access = False  # Don't track during initialization
            super().__init__(**kwargs)
            self._track_access = True  # Start tracking after initialization

        def __getattribute__(self, name):
            # No need to track access to private attributes or methods
            if not name.startswith("_"):
                # Use object.__getattribute__ to avoid infinite recursion
                track_access = object.__getattribute__(self, "_track_access")
                if track_access:
                    accessed_params = object.__getattribute__(self, "_accessed_params")
                    accessed_params.add(name)
            return object.__getattribute__(self, name)

        @property
        def accessed_params(self):
            return self._accessed_params

        def clear(self):
            self._accessed_params.clear()

    with mock_ctx, mock.patch.object(engine_class, "_check_unsupported_params"):
        remote_params = RemoteParams(api_url="test")
        if issubclass(engine_class, RemoteInferenceEngine):
            engine = engine_class(
                model_params=model_params, remote_params=remote_params
            )
        else:
            engine = engine_class(model_params)

        # Create config with tracking
        tracked_params = AccessTrackingGenerationParams()
        tracked_params.clear()

        inference_config = InferenceConfig(
            model=model_params, generation=tracked_params
        )

        if issubclass(engine_class, RemoteInferenceEngine):
            inference_config.remote_params = remote_params

            # To avoid running inference, we just call the method that converts
            # the conversation to the API input. This should access most of the
            # parameters.
            engine._convert_conversation_to_api_input(
                sample_conversation, tracked_params, model_params
            )
        elif engine_class == LlamaCppInferenceEngine:
            with patch.object(engine, "_llm") as mock_llm:
                mock_llm.create_chat_completion.return_value = {
                    "choices": [{"message": {"content": "test"}}]
                }

                engine.infer([sample_conversation], inference_config)
        elif engine_class == NativeTextInferenceEngine:
            inference_config.generation.exclude_prompt_from_response = False
            engine.infer([sample_conversation], inference_config)
        elif engine_class == VLLMInferenceEngine:
            with patch.object(engine, "_llm") as mock_vllm:
                mock_vllm.chat.return_value = [
                    mock.MagicMock(outputs=[mock.MagicMock(text="Some output")])
                ]
                engine.infer([sample_conversation], inference_config)
        else:
            engine.infer([sample_conversation], inference_config)

        # Get params that were supported but never accessed
        unused_params = engine.get_supported_params() - tracked_params.accessed_params

        assert not unused_params, (
            f"{engine_class.__name__} claims to support these parameters "
            f"but never accessed them: {unused_params}"
        )

        # Get params that were accessed but not marked as supported
        unregistered_params = (
            tracked_params.accessed_params - engine.get_supported_params()
        )
        unregistered_params.remove("accessed_params")  # Test param, ignore

        assert not unregistered_params, (
            f"{engine_class.__name__} accessed 'unsupported' parameters: "
            f"{unregistered_params}"
        )



import tempfile

import pytest

from oumi.utils.saver import (
    load_infer_prob,
    load_infer_prob_csv,
    save_infer_prob,
    save_infer_prob_csv,
)

TEMP_FILENAME = "temp_filename"
PROBABILITIES = [
    [
        [0.01, 0.02, 0.03, 0.04],
        [0.11, 0.12, 0.13, 0.14],
    ],
    [
        [1.01, 1.02, 1.03, 1.04],
        [1.11, 1.12, 1.13, 1.14],
    ],
]


@pytest.mark.parametrize(
    "num_batches,batch_size,save_fn,load_fn",
    [
        (1, 1, save_infer_prob, load_infer_prob),
        (1, 2, save_infer_prob, load_infer_prob),
        (2, 1, save_infer_prob, load_infer_prob),
        (2, 2, save_infer_prob, load_infer_prob),
        (1, 1, save_infer_prob_csv, load_infer_prob_csv),
        (1, 2, save_infer_prob_csv, load_infer_prob_csv),
        (2, 1, save_infer_prob_csv, load_infer_prob_csv),
        (2, 2, save_infer_prob_csv, load_infer_prob_csv),
    ],
)
def test_save_load_infer_probs(num_batches, batch_size, save_fn, load_fn):
    probabilities = []
    for batch_no in range(num_batches):
        batch_probabilities = []
        for batch_index in range(batch_size):
            batch_probabilities.append(PROBABILITIES[batch_no][batch_index])
        probabilities.append(batch_probabilities)

    with tempfile.TemporaryDirectory() as output_temp_dir:
        # Save probabilities to a file.
        save_file = f"{output_temp_dir}/{TEMP_FILENAME}"
        save_fn(save_file, probabilities)

        # Load probabilities from file and compare to the original ones.
        loaded_probabilities = load_fn(save_file)
        assert probabilities == loaded_probabilities



import os
import threading
import time
from pathlib import Path
from queue import Queue

import jsonlines
import pandas as pd
from openai import OpenAI
from tqdm import tqdm


def _get_model_name(model_id: str) -> str:
    segments = model_id.split("/")

    # Get first 5 characters of the snapshot id to keep the name small.
    snapshot = segments[-1][:5]
    model_segment_key = "models--"
    for s in segments:
        if model_segment_key in s:
            model_name = s[len(model_segment_key) :]
            return f"{model_name}_{snapshot}"
    # If the model_id is not in the HF format (ex. it's a path to a custom model),
    # return up to the last 20 characters.
    # We replace slashes in the directory path with underscores since the model
    # name is used as part of a filename.
    return model_id.replace("/", "_")[-20:]


def main() -> None:
    """Run inference against vLLM model hosted as an OpenAI API."""
    openai_api_key = "EMPTY"
    IP = os.environ["THIS_IP_ADDRESS"]
    openai_api_base = f"http://{IP}:8000/v1"
    client = OpenAI(
        # defaults to os.environ.get("OPENAI_API_KEY")
        api_key=openai_api_key,
        base_url=openai_api_base,
    )
    models = client.models.list()
    MODEL = models.data[0].id
    MODEL_NAME = _get_model_name(MODEL)
    JOB_NUMBER = os.environ["JOB_NUMBER"]
    INPUT_FILEPATH = Path(os.environ["OUMI_VLLM_INPUT_FILEPATH"])
    OUTPUT_DIR = Path(os.environ["OUMI_VLLM_OUTPUT_DIR"])
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    TIMESTR = time.strftime("%Y%m%d_%H%M%S")
    OUTPUT_FILENAME = f"{JOB_NUMBER}_vllm_output_{TIMESTR}_{MODEL_NAME}.jsonl"
    OUTPUT_FILEPATH = OUTPUT_DIR / OUTPUT_FILENAME
    print(f"Input filepath is {INPUT_FILEPATH}")
    print(f"Files will be output to {OUTPUT_FILEPATH}")

    if not INPUT_FILEPATH.is_file():
        raise FileNotFoundError(f"Input file not found: {INPUT_FILEPATH}")
    json_objects = pd.read_json(INPUT_FILEPATH, lines=True)
    all_messages = json_objects["messages"].to_list()
    write_queue = Queue()

    def _thread_write_to_file():
        while True:
            messages = write_queue.get()
            if messages is None:
                write_queue.task_done()
                break

            with jsonlines.open(OUTPUT_FILEPATH, mode="a") as writer:
                json_obj = {"messages": messages}
                writer.write(json_obj)
                write_queue.task_done()

    threading.Thread(target=_thread_write_to_file, daemon=True).start()
    for messages in tqdm(all_messages):
        chat_completion = client.chat.completions.create(
            messages=messages,
            model=MODEL,
        )

        messages.append(
            {"role": "assistant", "content": chat_completion.choices[0].message.content}
        )
        write_queue.put(messages)

    write_queue.put(None)
    write_queue.join()
    print("Inference complete")


if __name__ == "__main__":
    main()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional

from omegaconf import MISSING

from oumi.core.configs.base_config import BaseConfig
from oumi.core.configs.params.base_params import BaseParams


class DatasetSource(Enum):
    """Source of the dataset for analysis."""

    CONFIG = "config"
    """Load dataset from config parameters (dataset_name, dataset_path, etc.)"""
    DIRECT = "direct"
    """Pass dataset directly to DatasetAnalyzer.__init__()"""


@dataclass
class SampleAnalyzerParams(BaseParams):
    """Params for a single sample analyzer plugin."""

    id: str = MISSING
    """Unique identifier for the analyzer."""

    params: dict[str, Any] = field(default_factory=dict)
    """Analyzer-specific parameters passed to the analyzer constructor."""


@dataclass
class AnalyzeConfig(BaseConfig):
    """Configuration for dataset analysis and aggregation."""

    # Required field - must come first
    dataset_source: DatasetSource = MISSING
    """Source of the dataset for analysis. Use CONFIG to load from config parameters
    or DIRECT to pass dataset directly to DatasetAnalyzer.__init__().

    This field is required and must be explicitly set.
    """

    # Simple fields for common use cases
    dataset_name: Optional[str] = None
    """Dataset name."""

    dataset_path: Optional[str] = None
    """Path to a custom dataset file (JSON or JSONL format).
    If provided, this takes precedence over dataset_name for loading custom datasets.
    """

    dataset_format: Optional[str] = None
    """Format of the custom dataset. Either 'oumi' (conversation format) or 'alpaca'.
    Only used when dataset_path is provided.
    """

    split: str = "train"
    """The split of the dataset to load.
    This is typically one of "train", "test", or "validation". Defaults to "train".
    """

    subset: Optional[str] = None
    """The subset of the dataset to load. If None, uses the base dataset."""

    sample_count: Optional[int] = None
    """The number of examples to sample from the dataset.
    If None, uses the full dataset. If specified, must be non-negative.
    """

    output_path: str = "."
    """Directory path where output files will be saved.

    Defaults to current directory ('.').
    """

    analyzers: list[SampleAnalyzerParams] = field(default_factory=list)
    """List of analyzer configurations (plugin-style)."""

    tokenizer_config: Optional[dict[str, Any]] = None
    """Tokenizer configuration for building a tokenizer.
    If None, no tokenizer will be used.

    Expected format::

        {
            "model_name": "gpt2",  # Required: model name for tokenizer
            "tokenizer_kwargs": {},  # Optional: additional tokenizer parameters
            "trust_remote_code": False  # Optional: whether to trust remote code
        }

    """

    # Add processor parameters for vision-language datasets
    processor_name: Optional[str] = None
    """Processor name for vision-language datasets."""

    processor_kwargs: dict[str, Any] = field(default_factory=dict)
    """Processor-specific parameters."""

    trust_remote_code: bool = False
    """Whether to trust remote code for processor loading."""

    is_multimodal: Optional[bool] = None
    """If True, treat the dataset as multimodal (vision-language) when using a
    custom dataset_path. If False, treat as text-only.
    """

    def __post_init__(self):
        """Validates the configuration parameters."""
        if self.dataset_source == DatasetSource.CONFIG:
            # Only require dataset info when loading from config
            if not self.dataset_name and not self.dataset_path:
                raise ValueError(
                    "Either 'dataset_name' or 'dataset_path' must be provided when "
                    "dataset_source=DatasetSource.CONFIG"
                )
        else:
            # When using direct dataset, dataset_name is optional but recommended
            if not self.dataset_name:
                self.dataset_name = "Custom Dataset"

        # Validate dataset_format requirements
        if self.dataset_path is not None:
            if self.dataset_format is None:
                raise ValueError(
                    "'dataset_format' must be specified when using 'dataset_path'. "
                    "Use 'oumi' for conversation format or 'alpaca' for instruction "
                    "format."
                )
            elif self.dataset_format not in ["oumi", "alpaca"]:
                raise ValueError("'dataset_format' must be either 'oumi' or 'alpaca'")

            # Require explicit is_multimodal setting for custom datasets
            if self.is_multimodal is None:
                raise ValueError(
                    "'is_multimodal' must be specified when using 'dataset_path'. "
                    "Set to 'True' for vision-language datasets or 'False' for "
                    "text-only datasets."
                )

            # Additional validation for multimodal
            if self.is_multimodal is True:
                # Currently VLJsonlinesDataset expects oumi conversation format
                if self.dataset_format != "oumi":
                    raise ValueError(
                        "Multimodal datasets require dataset_format='oumi'"
                    )
                if not self.processor_name:
                    raise ValueError(
                        "'processor_name' must be specified when 'is_multimodal' "
                        "is True"
                    )

        # Validate sample_count
        if self.sample_count is not None and self.sample_count <= 0:
            raise ValueError("`sample_count` must be greater than 0.")

        # Validate analyzer configurations
        analyzer_ids = set()
        for analyzer in self.analyzers:
            # Validate analyzer ID
            if not analyzer.id:
                raise ValueError("Analyzer 'id' must be provided")
            if analyzer.id in analyzer_ids:
                raise ValueError(f"Duplicate analyzer ID found: '{analyzer.id}'")
            analyzer_ids.add(analyzer.id)



import os
from typing import Final, NamedTuple, Optional

import datasets
import pytest

from oumi.builders import (
    build_dataset_mixture,
    build_tokenizer,
)
from oumi.core.configs import (
    DataParams,
    DatasetParams,
    DatasetSplit,
    DatasetSplitParams,
    ModelParams,
)
from oumi.core.datasets import VisionLanguageSftDataset
from oumi.core.registry import REGISTRY, RegistryType

_DEFALT_DATASET_SPLIT: Final[str] = "test"
_DEFAULT_MODEL_NAME: Final[str] = "Qwen/Qwen2-VL-2B-Instruct"
_DEFAULT_CHAT_TEMPLATE: Final[str] = "qwen2-vl-instruct"


def _get_all_sft_vision_dataset_names() -> list[str]:
    """List all SFT datasets in the registry."""
    datasets = []
    for key, value in REGISTRY.get_all(RegistryType.DATASET).items():
        if issubclass(value, VisionLanguageSftDataset):
            datasets.append(key)
    return datasets


class LoadDatasetInfo(NamedTuple):
    dataset_name: str
    model_name: str
    stream: bool = False
    max_rows: int = 32
    expected_rows: Optional[int] = 32
    extra_dataset_features: Optional[list[str]] = None
    chat_template: str = _DEFAULT_CHAT_TEMPLATE
    dataset_subset: Optional[str] = None
    dataset_split: str = _DEFALT_DATASET_SPLIT
    collator_name: str = "vision_language_with_padding"
    trust_remote_code: bool = False


def get_dataset_test_id_fn(info):
    assert isinstance(info, LoadDatasetInfo), f"{type(info)}: {info}"
    return f"{info.dataset_name} {info.model_name}"


def _normalize_dataset_name_for_matching(s: str) -> str:
    return s.lower().strip()


def _get_all_sft_vision_dataset_infos() -> list[LoadDatasetInfo]:
    # Special case datasets that should be excluded from default testing.
    _EXCLUDED_DATASETS = set(
        {
            "coco_captions",
            "vision_language_jsonl",
            "vl_sft",
        }
    )

    all_dataset_names = set(_get_all_sft_vision_dataset_names())
    result = [
        LoadDatasetInfo(
            dataset_name="merve/vqav2-small",
            model_name=_DEFAULT_MODEL_NAME,
            dataset_split="validation",
            chat_template=_DEFAULT_CHAT_TEMPLATE,
            trust_remote_code=True,
            max_rows=64,
            expected_rows=64,
        ),
        LoadDatasetInfo(
            dataset_name="merve/vqav2-small",
            model_name="microsoft/Phi-3-vision-128k-instruct",
            dataset_split="validation",
            extra_dataset_features=["image_sizes"],
            chat_template="phi3-instruct",
            trust_remote_code=True,
            max_rows=32,
            expected_rows=32,
        ),
        LoadDatasetInfo(
            dataset_name="HuggingFaceM4/Docmatix",
            model_name=_DEFAULT_MODEL_NAME,
            dataset_subset="zero-shot-exp",
            dataset_split="test",
            chat_template=_DEFAULT_CHAT_TEMPLATE,
            trust_remote_code=True,
            max_rows=32,
            expected_rows=32,
        ),
        LoadDatasetInfo(
            dataset_name="HuggingFaceM4/the_cauldron",
            model_name=_DEFAULT_MODEL_NAME,
            dataset_subset="vqarad",
            dataset_split="train",
            chat_template=_DEFAULT_CHAT_TEMPLATE,
            trust_remote_code=True,
            max_rows=64,
            expected_rows=64,
        ),
        LoadDatasetInfo(
            dataset_name="allenai/pixmo-ask-model-anything",
            model_name=_DEFAULT_MODEL_NAME,
            dataset_split="train[10:20]",  # 404 error for some image URLs
            chat_template=_DEFAULT_CHAT_TEMPLATE,
            trust_remote_code=True,
            max_rows=64,
            expected_rows=None,
        ),
        LoadDatasetInfo(
            dataset_name="allenai/pixmo-cap",
            model_name=_DEFAULT_MODEL_NAME,
            dataset_split="train[50:51]",  # 429 error for some image URLs
            chat_template=_DEFAULT_CHAT_TEMPLATE,
            trust_remote_code=True,
            max_rows=64,
            expected_rows=None,
        ),
        LoadDatasetInfo(
            dataset_name="allenai/pixmo-cap-qa",
            model_name=_DEFAULT_MODEL_NAME,
            dataset_split="train[10:20]",  # 404 error for some image URLs
            chat_template=_DEFAULT_CHAT_TEMPLATE,
            trust_remote_code=True,
            max_rows=64,
            expected_rows=None,
        ),
    ]

    all_excluded_dataset_names_normalized = set(
        {
            _normalize_dataset_name_for_matching(name)
            for name in (
                _EXCLUDED_DATASETS.union({info.dataset_name for info in result})
            )
        }
    )

    for dataset_name in all_dataset_names:
        if (
            _normalize_dataset_name_for_matching(dataset_name)
            in all_excluded_dataset_names_normalized
        ):
            continue
        result.append(
            LoadDatasetInfo(
                dataset_name=dataset_name,
                model_name=_DEFAULT_MODEL_NAME,
                dataset_split=_DEFALT_DATASET_SPLIT,
                chat_template=_DEFAULT_CHAT_TEMPLATE,
                trust_remote_code=True,
            )
        )

    assert len(result) > 1
    for idx, info in enumerate(result):
        assert info.dataset_name, f"Index: {idx}"
        assert (
            _normalize_dataset_name_for_matching(info.dataset_name) in all_dataset_names
        ), f"Index: {idx}"
        assert info.model_name, f"Index: {idx}"
        assert info.chat_template, f"Index: {idx}"
        assert info.dataset_split, f"Index: {idx}"
        assert info.collator_name, f"Index: {idx}"

    return result


@pytest.mark.parametrize(
    "info", _get_all_sft_vision_dataset_infos(), ids=get_dataset_test_id_fn
)
@pytest.mark.e2e
def test_build_dataset_mixture(info: LoadDatasetInfo):
    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    debug_tag = f"Test: {info}"
    model_params = ModelParams(
        model_name=info.model_name,
        trust_remote_code=info.trust_remote_code,
        chat_template=info.chat_template,
    )
    tokenizer = build_tokenizer(model_params)
    train_split = DatasetSplitParams(
        collator_name=info.collator_name,
        stream=info.stream,
        datasets=[
            DatasetParams(
                dataset_name=info.dataset_name,
                subset=info.dataset_subset,
                split=info.dataset_split,
                shuffle=False,
                seed=42,
                trust_remote_code=info.trust_remote_code,
                dataset_kwargs={
                    "processor_name": info.model_name,
                    "limit": info.max_rows,
                    "return_tensors": True,
                },
            )
        ],
    )
    dataset = build_dataset_mixture(
        DataParams(train=train_split), tokenizer, DatasetSplit.TRAIN
    )

    assert isinstance(dataset, datasets.Dataset)

    assert dataset.num_rows > 0, debug_tag
    assert dataset.num_rows <= info.max_rows, debug_tag

    if info.expected_rows is not None:
        assert dataset.num_rows == info.expected_rows, debug_tag

    assert "input_ids" in dataset.features, debug_tag
    assert "attention_mask" in dataset.features, debug_tag
    assert "pixel_values" in dataset.features, debug_tag
    assert "labels" in dataset.features, debug_tag

    if info.extra_dataset_features is not None and len(info.extra_dataset_features) > 0:
        for extra_feature in info.extra_dataset_features:
            assert extra_feature in dataset.features, debug_tag

    assert dataset[0] is not None, debug_tag
    assert dataset[dataset.num_rows - 1] is not None, debug_tag



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from typing import Any, Optional

import pydantic
from typing_extensions import override

from oumi.core.configs import GenerationParams, ModelParams, RemoteParams
from oumi.core.configs.params.guided_decoding_params import GuidedDecodingParams
from oumi.core.types.conversation import Conversation
from oumi.inference.remote_inference_engine import RemoteInferenceEngine


class GoogleVertexInferenceEngine(RemoteInferenceEngine):
    """Engine for running inference against Google Vertex AI."""

    _API_URL_TEMPLATE = (
        "https://{region}-aiplatform.googleapis.com/v1beta1/projects/"
        "{project_id}/locations/{region}/endpoints/openapi/chat/completions"
    )
    """The API URL template for the GCP project. Used when no `api_url` is provided."""

    _DEFAULT_PROJECT_ID_ENV_KEY: str = "PROJECT_ID"
    """The default project ID environment key for the GCP project."""

    _DEFAULT_REGION_ENV_KEY: str = "REGION"
    """The default region environment key for the GCP project."""

    _project_id: Optional[str] = None
    """The project ID for the GCP project."""

    _region: Optional[str] = None
    """The region for the GCP project."""

    def __init__(
        self,
        model_params: ModelParams,
        *,
        generation_params: Optional[GenerationParams] = None,
        remote_params: Optional[RemoteParams] = None,
        project_id_env_key: Optional[str] = None,
        region_env_key: Optional[str] = None,
        project_id: Optional[str] = None,
        region: Optional[str] = None,
    ):
        """Initializes the inference Engine.

        Args:
            model_params: The model parameters to use for inference.
            generation_params: The generation parameters to use for inference.
            remote_params: The remote parameters to use for inference.
            project_id_env_key: The environment variable key name for the project ID.
            region_env_key: The environment variable key name for the region.
            project_id: The project ID to use for inference.
            region: The region to use for inference.
        """
        super().__init__(
            model_params=model_params,
            generation_params=generation_params,
            remote_params=remote_params,
        )
        if project_id and project_id_env_key:
            raise ValueError(
                "You cannot set both `project_id` and `project_id_env_key`."
            )
        if region and region_env_key:
            raise ValueError("You cannot set both `region` and `region_env_key`.")

        self._project_id_env_key = (
            project_id_env_key or self._DEFAULT_PROJECT_ID_ENV_KEY
        )
        self._region_env_key = region_env_key or self._DEFAULT_REGION_ENV_KEY
        self._project_id = project_id
        self._region = region

    @override
    def _set_required_fields_for_inference(self, remote_params: RemoteParams) -> None:
        """Set required fields for inference."""
        if (
            not remote_params.api_url
            and not self._remote_params.api_url
            and not self.base_url
        ):
            if self._project_id and self._region:
                project_id = self._project_id
                region = self._region
            elif os.getenv(self._project_id_env_key) and os.getenv(
                self._region_env_key
            ):
                project_id = os.getenv(self._project_id_env_key)
                region = os.getenv(self._region_env_key)
            else:
                raise ValueError(
                    "This inference engine requires that either `api_url` is set in "
                    "`RemoteParams` or that both `project_id` and `region` are set. "
                    "You can set the `project_id` and `region` when "
                    "constructing a GoogleVertexInferenceEngine, "
                    f"or as environment variables: `{self._project_id_env_key}` and "
                    f"`{self._region_env_key}`."
                )

            remote_params.api_url = self._API_URL_TEMPLATE.format(
                project_id=project_id,
                region=region,
            )

        super()._set_required_fields_for_inference(remote_params)

    @override
    def _get_api_key(self, remote_params: RemoteParams) -> str:
        """Gets the authentication token for GCP."""
        try:
            from google.auth import default  # pyright: ignore[reportMissingImports]
            from google.auth.transport.requests import (  # pyright: ignore[reportMissingImports]
                Request,
            )
            from google.oauth2 import (  # pyright: ignore[reportMissingImports]
                service_account,
            )
        except ModuleNotFoundError:
            raise RuntimeError(
                "Google-auth is not installed. "
                "Please install oumi with GCP extra:`pip install oumi[gcp]`, "
                "or install google-auth with `pip install google-auth`."
            )

        if remote_params.api_key:
            credentials = service_account.Credentials.from_service_account_file(
                filename=remote_params.api_key,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        else:
            credentials, _ = default(
                scopes=["https://www.googleapis.com/auth/cloud-platform"]
            )
        credentials.refresh(Request())  # type: ignore
        return credentials.token  # type: ignore

    @override
    def _get_request_headers(
        self, remote_params: Optional[RemoteParams]
    ) -> dict[str, str]:
        """Gets the request headers for GCP."""
        if not remote_params:
            raise ValueError("Remote params are required for GCP inference.")

        headers = {
            "Authorization": f"Bearer {self._get_api_key(remote_params)}",
            "Content-Type": "application/json",
        }
        return headers

    @override
    def _default_remote_params(self) -> RemoteParams:
        """Returns the default remote parameters."""
        return RemoteParams(num_workers=10, politeness_policy=60.0)

    @override
    def _convert_conversation_to_api_input(
        self,
        conversation: Conversation,
        generation_params: GenerationParams,
        model_params: ModelParams,
    ) -> dict[str, Any]:
        """Converts a conversation to an OpenAI input.

        Documentation: https://cloud.google.com/vertex-ai/generative-ai/docs/multimodal/call-vertex-using-openai-library

        Args:
            conversation: The conversation to convert.
            generation_params: Parameters for generation during inference.
            model_params: Model parameters to use during inference.

        Returns:
            Dict[str, Any]: A dictionary representing the Vertex input.
        """
        api_input = {
            "model": model_params.model_name,
            "messages": self._get_list_of_message_json_dicts(
                conversation.messages, group_adjacent_same_role_turns=True
            ),
            "max_completion_tokens": generation_params.max_new_tokens,
            "temperature": generation_params.temperature,
            "top_p": generation_params.top_p,
            "n": 1,  # Number of completions to generate for each prompt.
            "seed": generation_params.seed,
            "logit_bias": generation_params.logit_bias,
        }

        if generation_params.stop_strings:
            api_input["stop"] = generation_params.stop_strings

        if generation_params.guided_decoding:
            api_input["response_format"] = _convert_guided_decoding_config_to_api_input(
                generation_params.guided_decoding
            )

        return api_input

    @override
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine."""
        return {
            "guided_decoding",
            "logit_bias",
            "max_new_tokens",
            "seed",
            "stop_strings",
            "temperature",
            "top_p",
        }


#
# Helper functions
#
def _convert_guided_decoding_config_to_api_input(
    guided_config: GuidedDecodingParams,
) -> dict:
    """Converts a guided decoding configuration to an API input."""
    if guided_config.json is None:
        raise ValueError(
            "Only JSON schema guided decoding is supported, got '%s'",
            guided_config,
        )

    json_schema = guided_config.json

    if isinstance(json_schema, type) and issubclass(json_schema, pydantic.BaseModel):
        schema_name = json_schema.__name__
        schema_value = json_schema.model_json_schema()
    elif isinstance(json_schema, dict):
        # Use a generic name if no schema is provided.
        schema_name = "Response"
        schema_value = json_schema
    elif isinstance(json_schema, str):
        # Use a generic name if no schema is provided.
        schema_name = "Response"
        # Try to parse as JSON string
        schema_value = json.loads(json_schema)
    else:
        raise ValueError(
            f"Got unsupported JSON schema type: {type(json_schema)}"
            "Please provide a Pydantic model or a JSON schema as a "
            "string or dict."
        )

    return {
        "type": "json_schema",
        "json_schema": {
            "name": schema_name,
            "schema": _replace_refs_in_schema(schema_value),
        },
    }


def _replace_refs_in_schema(schema: dict) -> dict:
    """Replace $ref references in a JSON schema with their actual definitions.

    Args:
        schema: The JSON schema dictionary

    Returns:
        dict: Schema with all references replaced by their definitions and $defs removed
    """

    def _get_ref_value(ref: str) -> dict:
        # Remove the '#/' prefix and split into parts
        parts = ref.replace("#/", "").split("/")

        # Navigate through the schema to get the referenced value
        current = schema
        for part in parts:
            current = current[part]
        return current.copy()  # Return a copy to avoid modifying the original

    def _replace_refs(obj: dict) -> dict:
        if not isinstance(obj, dict):
            return obj

        result = {}
        for key, value in obj.items():
            if key == "$ref":
                # If we find a $ref, replace it with the actual value
                return _replace_refs(_get_ref_value(value))
            elif isinstance(value, dict):
                result[key] = _replace_refs(value)
            elif isinstance(value, list):
                result[key] = [
                    _replace_refs(item) if isinstance(item, dict) else item
                    for item in value
                ]
            else:
                result[key] = value

        return result

    # Replace all references first
    resolved = _replace_refs(schema.copy())

    # Remove the $defs key if it exists
    if "$defs" in resolved:
        del resolved["$defs"]

    return resolved
